#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from odoo.tests.common import TransactionCase

Hours = {"15:30": 15.5, "12:00": 12, "9:45": 9.75}


class TestTimeRange(TransactionCase):
    def setUp(self):
        super(TestTimeRange, self).setUp()
        self.Model = self.env["test.time.range"]

    def test_compute_range(self):
        obj = self.Model.create(dict(time_value=Hours["15:30"]))

        self.assertEqual(obj.range_value, "afternoon")

        obj.time_value = Hours["9:45"]

        self.assertEqual(obj.range_value, "morning")

        obj.time_value = Hours["12:00"]

        self.assertEqual(obj.range_value, "noon")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf.odoo import api, fields, models, _


def get_mixin_descendants(pool, mixin):
    """ Get the models that inherit from `mixin`.

    :param mixin: mixin name to get it descendants.

    """
    for model_name in pool.descendants([mixin], "_inherit"):
        if not pool[model_name]._abstract:
            yield model_name


class TypedReference(fields.Reference):
    """A reference field filtered by a type (mixin).

    This is like a Many2one field, but the comodel should be an AbstractModel
    (a mixin).  This way you can relate objects in different actual models but
    that share a common type.

    .. warning:: This field works only in Odoo 11+.

    """

    type = "reference"
    _slots = {"mixin": None, "comodel_name": None, "delegate": None}

    def __init__(self, mixin=fields.Default, delegate=fields.Default, **kwargs):
        # Set comodel_name = mixin, this is required for odoo make triggers
        # on delegate field declarations.
        super(TypedReference, self).__init__(
            **dict(kwargs, mixin=mixin, comodel_name=mixin, delegate=delegate)
        )

    def new(self, **kwargs):
        # Pass original args to the new one.  This ensures that the
        # mixin arg is present.  In odoo/models.py, Odoo calls
        # this `new()` without arguments to duplicate the fields from parent
        # classes.
        kwargs = dict(self.args, **kwargs)
        return super(TypedReference, self).new(**kwargs)

    def _setup_regular_base(self, model):
        if not self.selection:

            def selection(model):
                return [
                    (model_name, _(model.env[model_name]._description or model_name))
                    for model_name in get_mixin_descendants(model.pool, self.mixin)
                ]

            self.selection = selection
        return super(TypedReference, self)._setup_regular_base(model)

    def convert_to_cache(self, value, record, validate=True):
        res = super(TypedReference, self).convert_to_cache(
            value, record, validate=validate
        )
        if res and validate:
            res_model, res_id = res
            descendants = [
                model_name
                for model_name in get_mixin_descendants(record.pool, self.mixin)
            ]
            if res_model not in descendants:
                raise ValueError(_("Wrong value for %s: %r") % (self, res))
        return res


@api.model
def _setup_base(self, *args, **kwargs):
    super_setup_base(self, *args, **kwargs)
    delegate_reference_fields = {
        field.mixin: (name, field)
        for name, field in self._fields.items()
        if isinstance(field, TypedReference) and field.delegate
    }
    self._add_inherits_by_reference_fields(delegate_reference_fields.values())


@api.model
def _add_inherits_by_reference_fields(self, reference_fields):
    """ Determine inherited fields. """
    # determine candidate inherited fields
    for reference_field_name, reference_field in reference_fields:
        mixin = self.env[reference_field.mixin]
        for name, field in mixin._fields.items():
            if name not in self._fields:
                args = dict(
                    field.args,
                    compute=_make_compute_method(reference_field_name, name),
                    depends=(
                        reference_field_name,
                        "%s.%s" % (reference_field_name, name),
                    ),
                )
                # determine dependencies, compute, inverse, and search
                if not (field.readonly or reference_field.readonly):
                    args["inverse"] = _make_inverse_method(reference_field_name, name)
                if field._description_searchable:
                    # allow searching on self only if the related
                    # field is searchable
                    args["search"] = _make_search_method(reference_field_name, name)
                self._add_field(name, field.new(**args))


def _make_compute_method(reference_field, field_name):
    @api.multi
    def _get(self):
        for record in self:
            reference = record[reference_field]
            if reference:
                record[field_name] = reference[field_name]

    return _get


def _make_inverse_method(reference_field, field_name):
    @api.multi
    def _set(self):
        for record in self:
            reference = record[reference_field]
            if reference:
                reference[field_name] = record[field_name]
            else:
                raise ValueError("TODO")

    return _set


def _make_search_method(reference_field, field_name):
    @api.multi
    def _search(self, operator, value):
        return [("%s.%s" % (reference_field, field_name), operator, value)]

    return _search


super_setup_base = models.BaseModel._setup_base
models.BaseModel._setup_base = _setup_base
models.BaseModel._add_inherits_by_reference_fields = _add_inherits_by_reference_fields



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""External OpenERP's addons

"""
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import sys
import logging
import re

from xoutil.future.functools import lru_cache
from xoutil.modules import customize
from xoutil.modules import modulemethod
from xoutil.string import cut_prefix

from xoeuf.eight import string_types


# In Odoo 10, they allow to import from both 'odoo' and 'openerp'
_ADDONS_NAMESPACE = re.compile(r"^(?:odoo|openerp)\.addons\.(?P<module>[^\.]+)\.")

XOEUF_EXTERNAL_ADDON_GROUP = "xoeuf.addons"


# XXX: @manu, probably the prefix 'xoeuf.' could be avoided.
class OdooHook(object):
    """Hook for 'odoo' (or 'openerp') package to be available as 'xoeuf.odoo'.

    """

    try:
        import openerp as _mod
    except ImportError:
        # In Odoo 9 they have an 'odoo.py' that is importable when developing
        # (buildout, etc), so we have to try to import 'openerp' before trying
        # 'odoo'.
        import odoo as _mod
    NAME = _mod.__name__
    del _mod
    REGEX = r"^xoeuf[.](?:openerp|odoo)\b"

    def find_module(self, name, path=None):
        import re

        if re.match(self.REGEX, name):
            return self

    def load_module(self, name):
        import sys
        import re
        import importlib

        assert name not in sys.modules
        regex = self.REGEX + r"(.*)"
        canonical = re.sub(regex, self.NAME + r"\g<1>", name)
        if canonical in sys.modules:
            mod = sys.modules[canonical]
        else:
            # probable failure
            mod = importlib.import_module(canonical)
        # just set the original module at the new location. Don't proxy,
        # it breaks *-import (unless you can find how `from a import *` lists
        # what's supposed to be imported by `*`, and manage to override it)
        sys.modules[name] = mod
        return sys.modules[name]


sys.meta_path.insert(0, OdooHook())


class _PatchesRegistry(object):
    _registry = {}
    _wrapped = {}

    def __call__(self, func):
        from xoutil.names import nameof

        name = nameof(func, inner=True, full=False)
        self._registry[name] = func
        return func

    def get_super(self, name):
        return self._wrapped[name]

    def apply(self):
        from xoeuf.odoo.modules import module

        patched = getattr(module, "__xoeuf_patched__", False)
        if patched:
            # This is an Odoo that's being patched by us.
            return
        bootstraped = getattr(self, "bootstraped", False)
        if not bootstraped:
            for name, _func in self._registry.items():
                self._wrapped[name] = getattr(module, name)
            module = customize(module)[0]
            for name, func in self._registry.items():
                setattr(module, name, func)
            self.bootstraped = True


patch = _PatchesRegistry()


@modulemethod
@lru_cache(1)
def find_external_addons(self):
    """Finds all externally installed addons.

    Externally installed addons are modules that are distributed with
    setuptools' distributions.

    An externally addon is defined in a package that defines an `entry
    point`__ in the group "xoeuf.addons" which points to a standard package
    (i.e loadable without any specific loader).

    :returns:  A dictionary from addons to it's a tuple of `(container's path,
               entry_point)`.

    Example::

       [xoeuf.addons]
       xopgi_account = xopgi.addons.xopgi_account

    """
    import os
    from pkg_resources import iter_entry_points
    from xoutil.future.itertools import delete_duplicates

    res = []
    for entry in iter_entry_points(XOEUF_EXTERNAL_ADDON_GROUP):
        if not entry.attrs:
            # The entry-point is a whole module.  We can't load the module
            # here, cause the whole point is to grab the paths before openerp
            # is configured, but if you load an OpenERP addon you will be
            # importing openerp somehow and enacting configuration
            loc = entry.dist.location
            relpath = entry.module_name.replace(".", os.path.sep)
            # The parent directory is the one!
            abspath = os.path.abspath(os.path.join(loc, relpath, ".."))
            if os.path.isdir(abspath):
                res.append(abspath)
                name = entry.module_name
                pos = name.rfind(".")
                if pos >= 0:
                    name = name[pos + 1 :]  # noqa
    return delete_duplicates(res)


@patch
@modulemethod
def initialize_sys_path(self):
    from xoutil.objects import setdefaultattr
    from xoeuf.odoo.modules import module

    _super = patch.get_super("initialize_sys_path")
    external_addons = setdefaultattr(self, "__addons", [])
    if not external_addons:
        _super()
        result = module.ad_paths
        external_addons.extend(self.find_external_addons())
        result.extend(external_addons)
        module.ad_paths = result
        return result
    else:
        return module.ad_paths


def patch_modules():
    """Patches OpenERP `modules.module` to work with external addons.

    """
    patch.apply()


def _get_registry(db_name):
    """Helper method to get the registry for a `db_name`."""
    from odoo.modules.registry import Registry

    if isinstance(db_name, string_types):
        db = Registry(db_name)
    elif isinstance(db_name, Registry):
        db = db_name
    else:
        import sys

        caller = sys.getframe(1).f_code.co_name
        raise TypeError('"%s" requires a string or a Registry' % caller)
    return db


def get_dangling_modules(db):
    """Get registered modules that are no longer available.

    Returns the recordset of dangling modules.  This is a recordset of the
    model `ir.module.module`.

    A dangling module is one that is listed in the DB, but is not reachable in
    any of the addons paths (not even externally installed).

    :param db: Either the name of the database to load or a `registry
               <xoeuf.odoo.modules.registry.Registry>`:class:.

    :return: A record-set with dangling modules.

    .. warning:: We create a new cursor to the DB and the returned recordset
                 uses it.

    """
    from xoeuf import api
    from xoeuf.odoo import SUPERUSER_ID
    from xoeuf.odoo.modules.module import get_modules

    registry = _get_registry(db)
    cr = registry.cursor()
    env = api.Environment(cr, SUPERUSER_ID, {})
    Module = env["ir.module.module"]
    available = get_modules()
    return Module.search([("name", "not in", available)])


def mark_dangling_modules(db):
    """Mark `dangling <get_dangling_modules>`:func: as uninstallable.

    Parameters and return value are the same as in function
    :func:`get_dangling_modules`.

    """
    dangling = get_dangling_modules(db)
    dangling.write(dict(state="uninstallable"))
    dangling.env.cr.commit()
    return dangling


def get_object_module(obj, typed=False):
    """Return the name of the OpenERP addon the `obj` has been defined.

    If the `obj` is not defined (imported) from the "openerp.addons."
    namespace, return None.

    """
    from xoutil.names import nameof

    name = nameof(obj, inner=True, full=True, typed=typed)
    match = _ADDONS_NAMESPACE.match(name)
    if match:
        module = match.groupdict()["module"]
        return module
    else:
        return None


def is_object_installed(self, object):
    """Detects if `object` is installed in the DB.

    `self` must be an Odoo model (recordset, but it may be empty).

    """
    module = get_object_module(object)
    if module:
        mm = self.env["ir.module.module"].sudo()
        query = [("state", "=", "installed"), ("name", "=", module)]
        return bool(mm.search(query))
    else:
        return False


def get_caller_addon(depth=0, max_depth=5):
    """Guess the caller addon.

    :param depth: Skip that many levels in the call stack.

    :param max_depth: Max level to look in the call stack.

    Technically, we look in the globals of the *calling* stack frame for the
    ``__name__`` and, its matches the format with 'oddo.addons.<addon>.*',
    return the addon name; otherwise, look up in the frame stack until
    `max_depth` is reached.  No addon name is found, return None.

    .. versionchanged:: 0.51.0 Added `max_depth` argument.

    """
    res = None
    frame = sys._getframe(1 + depth)
    while depth < max_depth and frame is not None and not res:
        module = frame.f_globals["__name__"]
        if module.startswith("odoo.addons."):
            module = cut_prefix(module, "odoo.addons.")
            res = module.split(".", 1)[0]
        elif module.startswith("openerp.addons."):
            module = cut_prefix(module, "openerp.addons.")
            res = module.split(".", 1)[0]
        depth += 1
        frame = frame.f_back
    return res


del re, logging, OdooHook



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Part of xouti.eight we need to keep Python 2.7 compatibility.

"""
import sys

_py2 = sys.version_info[0] == 2
_py3 = sys.version_info[0] == 3

try:
    input = raw_input
except NameError:
    input = input

try:
    from xoutil.eight import string_types
except ImportError:
    string_types = (str,)

try:
    from xoutil.eight import integer_types
except ImportError:
    integer_types = (int,)


try:
    import __builtin__  # noqa

    builtins = __builtin__
except ImportError:
    import builtins  # noqa

    __builtin__ = builtins


try:
    exec_ = getattr(builtins, "exec")  # noqa
except AttributeError:

    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        import sys

        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")


def typeof(obj):
    """Obtain the object's type (compatible with Python 2**3)."""
    if _py3:
        return type(obj)
    else:
        from types import InstanceType

        return obj.__class__ if isinstance(obj, InstanceType) else type(obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf.odoo.tests.common import TransactionCase
from xoeuf.models.extensions import get_ref


class TestModels(TransactionCase):
    def test_get_ref(self):
        Foo = self.env["test_xoeuf_models.foobar"]
        record = self.env.ref("test_models.record1")
        self.assertEqual(record, get_ref(Foo, "record1"))
        self.assertEqual(record, Foo.get_ref("record1"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from collections import namedtuple
from xoutil.string import cut_prefix

from xoeuf import models, api
from xoeuf.eight import string_types

from odoo import fields

__all__ = ["Enumeration"]

Member = namedtuple("Member", ("name", "value"))


def Enumeration(enumclass, *args, **kwargs):
    """Create an enumeration field.

    The `enumclass` argument must be either an `enumeration class` or a
    fully qualified name of an enumeration class.

    The column in the DB will be of type CHAR and the values are the name of
    attribute in the enumeration class.

    Enumeration classes are required to:

    - have an attribute ``__members__`` with a `collections.Mapping`:class:
      from names to values.

    - have an attribute for each name in ``__member__`` with the same value as
      in the mapping.

    Those rules imply::

       >>> all(getattr(enumclass, name) is value
       ...     for name, value in enumclass.__members__.items())

    These requirements are compatible with Python 3.4's `enum.Enum`:class:
    However, we don't require that values are instances of the enumeration
    class.

    .. rubric:: Automatic selection field

    Enumeration fields cannot be put in views because their values can be
    arbitrary Python objects which are not easily transferred to/from the web
    client.

    You can use the parameter 'selection_field_name' to automatically inject a
    computed Selection field that allows to represent this Enumeration in the
    web client.  This automatic is by default not stored in the DB.  Its
    values are the names of the members of the enumeration class.

    You may pass a function 'compute_member_string' which will be called to
    get the user facing text of each member.  The function will be passed both
    the name and value of the member.  By default we return the member's name.

    Additional keyword arguments starting with the prefix ``selection_field_``
    are used to construct the Selection field.  For instance,
    'selection_field_string' sets the argument 'string'.  If not provided, we
    set 'string', and 'help' to the same values of the Enumeration field.

    .. versionchanged:: 0.36.0 A new generalized enumeration field.  Maintains
       DB compatibility (does not need migrations), but do require changes in
       the code.

    .. versionchanged:: 0.47.0 Add keyword parameter 'force_char_column'.

    .. versionchanged:: 0.61.0 The DB column type is always a CHAR.  Ignore
       the parameter 'force_char_column'.

    .. versionchanged:: 0.61.0 Add an automatic way to create a selection
       field via the parameter 'selection_field_name'.

    """
    from odoo.fields import Char
    from xoutil.objects import import_object, classproperty

    kwargs.pop("force_char_column", False)
    enumclass = import_object(enumclass)

    SELECTION_FIELD_PREFIX = "selection_field_"
    selection_field_kwargs = {
        cut_prefix(kw, SELECTION_FIELD_PREFIX): value
        for kw, value in kwargs.items()
        if kw.startswith(SELECTION_FIELD_PREFIX)
    }
    compute_member_string = kwargs.pop("compute_member_string", None)
    if compute_member_string:
        selection_field_kwargs["compute_member_string"] = compute_member_string

    class EnumeratedField(Char, _EnumeratedField):
        @classproperty
        def members(cls):
            return enumclass.__members__

        def setup_full(self, model):
            # Injects a new class in the model MRO, so that we can guarantee
            # the expectation about the methods create, write and search.
            #
            # The alternative would be to implement pre_save and pre_search
            # receivers.  But that would run for ALL models regardless of the
            # presence of Enumeration fields.  Also, we need to rewrite the
            # argument in the 'search' method, and signals are designed for
            # such use case (although they work at the moment).
            #
            # We must find EnumerationAdapter in the MRO to avoid injecting it
            # twice when using Enumeration fields in abstract models.
            #
            # We don't need to inject the EnumerationAdapter in related
            # (delegated) copies for this Enumeration fields, because the
            # actual field does the right thing with DB (it do have the
            # EnumerationAdapter injected.)
            cls = type(model)
            if EnumerationAdapter not in cls.mro() and not self.related:
                cls.__bases__ = (EnumerationAdapter,) + cls.__bases__
                # I tried to use model._add_field because it changes the
                # cls._fields that Odoo is iterating and that raises an error
                # (RuntimeError: dictionary changed size during iteration).
                #
                # So let's resort to injection.
                if selection_field_kwargs:
                    selection_field_name = selection_field_kwargs.pop("name")
                    selection_field = self.get_selection_field(
                        self.name, selection_field_name, **selection_field_kwargs
                    )
                    cls.__bases__ = (
                        SelectionMixin(
                            self.name, selection_field_name, selection_field
                        ),
                    ) + cls.__bases__
            return super(EnumeratedField, self).setup_full(model)

        def convert_to_read(self, value, record, use_name_get=True):
            if value is not None and value is not False:
                return self.get_member_by_value(value).value
            else:
                return value

        def convert_to_write(self, value, record):
            if value is not None and value is not False:
                if value in enumclass.__members__.values():
                    member = self.get_member_by_value(value)
                    # Our EnumerationAdapter takes care of doing the right
                    # thing when writing to the DB, also convert_to_column
                    # does.
                    return member.value
            return value

        def convert_to_cache(self, value, record, validate=True):
            if value not in enumclass.__members__.values():
                if value is not None and value is not False:
                    return self.get_member_by_name(value).value
            return value

        # NOTE: The parameter `values` was introduced in Odoo 11; the
        # parameter validate was introduced in Odoo 12.  We don't use
        # either; but declare them both to work across the three Odoo
        # versions.
        def convert_to_column(self, value, record, values=None, validate=True):
            if value in enumclass.__members__.values():
                return Char.convert_to_column(
                    self, self.get_member_by_value(value).name, record
                )
            else:
                return Char.convert_to_column(self, value, record)

    return EnumeratedField(*args, **kwargs)


class Adapter(object):
    # See the note in setup_full above.

    # Odoo 11 requires all bases to have these attributes.  See
    # odoo/models.py, method _build_model_attributes in lines 515-555 (at the
    # time of writing).
    _table = None
    _sequence = None
    _inherits = _depends = {}
    _sql_constraints = []
    _constraints = []
    _description = None

    # Odoo 12 requires this attribute.  See odoo/models.py, line 528
    _inherit = None


class EnumerationAdapter(Adapter):
    "Adapt the create/write/search method to Enumeration fields."

    @api.model
    @api.returns(*models.BaseModel.search._returns)
    def search(self, args, *pos_args, **kwargs):
        for index, query_part in enumerate(args):
            if not isinstance(query_part, string_types):
                fieldname, operator, operands = query_part
                field = self._fields.get(fieldname, None)
                if isinstance(field, _EnumeratedField):
                    if operator in ("=", "!="):
                        values = _get_db_value(field, operands)
                    elif operator in ("in", "not in"):
                        values = [_get_db_value(field, o) for o in operands]
                    else:
                        raise TypeError(
                            "Unsupported operator %r for an enumeration field"
                            % operator
                        )
                    args[index] = (fieldname, operator, values)
        return super(EnumerationAdapter, self).search(args, *pos_args, **kwargs)

    @api.model
    @api.returns("self", lambda value: value.id)
    def create(self, values):
        for fieldname, value in dict(values).items():
            field = self._fields.get(fieldname, None)
            values[fieldname] = _get_db_value(field, value)
        return super(EnumerationAdapter, self).create(values)

    @api.multi
    def write(self, values):
        for fieldname, value in dict(values).items():
            field = self._fields.get(fieldname, None)
            values[fieldname] = _get_db_value(field, value)
        return super(EnumerationAdapter, self).write(values)


def SelectionMixin(field_name, selection_field_name, selection_field):
    return type(
        "selection_mixin_" + field_name,
        (Adapter,),
        {selection_field_name: selection_field},
    )


class _EnumeratedField(object):
    def get_selection_field(
        self, field_name, name, compute_member_string=None, **kwargs
    ):
        """Return a computed Selection field to set/get the Enumeration.

        """
        selection_field_name = name

        @api.multi
        @api.depends(field_name)
        def _compute_selection_field(rs):
            for record in rs:
                value = getattr(record, field_name, None)
                if value is not None:
                    member = self.get_member_by_value(value)
                    setattr(record, selection_field_name, member.name)
                else:
                    setattr(record, selection_field_name, False)

        @api.multi
        def _set_selection(rs):
            for record in rs:
                key = getattr(record, selection_field_name, None)
                if key:
                    member = self.get_member_by_name(key)
                    setattr(record, field_name, member.value)
                else:
                    setattr(record, field_name, False)

        if not compute_member_string:
            compute_member_string = self._compute_member_string
        kwargs.setdefault("store", False)
        kwargs.setdefault("compute", _compute_selection_field)
        kwargs.setdefault("inverse", _set_selection)
        return fields.Selection(
            selection=lambda s: [
                (name, compute_member_string(name, value))
                for name, value in self.members.items()
            ],
            **kwargs
        )

    @staticmethod
    def _compute_member_string(name, value):
        return name

    @classmethod
    def get_member_by_value(cls, value):
        """Find the enumclass's member that is equal to `value`."""
        try:
            return next(Member(k, v) for k, v in cls.members.items() if v == value)
        except StopIteration:
            raise ValueError(
                "Invalid member %r of enumeration %r" % (value, cls.members)
            )

    @classmethod
    def get_member_by_name(cls, name):
        """Find the enumclass's member by name"""
        try:
            return Member(name, cls.members[name])
        except (AttributeError, KeyError):
            raise ValueError("Invalid key %r of enumeration %r" % (name, cls.members))


def _get_db_value(field, value):
    if value is None or value is False:  # and not field.required
        return value
    if isinstance(field, _EnumeratedField):
        member = field.get_member_by_value(value)
        return member.name
    else:
        return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

# TODO: @manu, At least in Odoo 10, there is 'openerp.sentrylog'
try:
    from openerp.sentrylog import *  # noqa: reexport
except ImportError:
    from ._sentrylog import *  # noqa: reexport



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf import models, fields


class Line(models.Model):
    _name = "test.monetary.line"
    value = fields.Monetary()
    currency_id = fields.Many2one("res.currency")


class ConcreteLine(models.Model):
    _name = "test.monetary.concrete"
    value = fields.Monetary(concrete=True)
    currency_id = fields.Many2one("res.currency")


class RelatedLine(models.Model):
    _name = "test.monetary.related"
    value = fields.Monetary(concrete=True)
    company_id = fields.Many2one("res.company")
    currency_id = fields.Many2one(related="company_id.currency_id")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf import api, fields, models


class APIModel(models.Model):
    _name = "xoeuf.tests.test_api.model"

    @api.from_active_ids(leak_context=True)
    def return_self_ids(self):
        return list(self.ids)

    @api.from_active_ids
    def return_ids_and_call_method(self, ids, methodname):
        return list(self.ids), getattr(self.browse(ids), methodname)()

    @api.from_active_ids(leak_context=True)
    def leaked_return_ids_and_call_method(self, ids, methodname):
        return list(self.ids), getattr(self.browse(ids), methodname)()


class Users(models.Model):
    _inherit = "res.users"

    text_field = fields.Char()

    # 'name' and 'partner_id.name' are the same thing.
    @api.onupdate("partner_id", "name", "partner_id.name")
    def update_text_field(self):
        for record in self:
            record.text_field = record.get_text_field()

    @api.requires_singleton
    def get_text_field(self):
        return "s2 %s" % self.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from enum import IntEnum, Enum
from xoeuf import models, fields


class COLORS(IntEnum):
    Blue = 0
    Red = 1
    Green = 2


class CARS(Enum):
    FORD = object()
    CHEV = object()


class Pax(object):
    single = 1
    double = 2
    __members__ = {"single": single, "double": double}


class WorkType(int):
    pass


class WORK_TYPE(object):
    easy = WorkType(0)
    hard = WorkType(1)
    __members__ = {"easy": easy, "hard": hard}


class Mixin(models.AbstractModel):
    _name = "test_enumeration.mixin"
    color = fields.Enumeration(
        COLORS, default=COLORS.Red, selection_field_name="color_name"
    )


class Model(models.Model):
    _name = "test.enum.model"
    _inherit = ["test_enumeration.mixin"]

    car = fields.Enumeration(CARS)
    pax = fields.Enumeration(Pax)


class DelegatedModel(models.Model):
    _name = "test.enum.model_delegated"
    _inherits = {"test.enum.model": "model_id"}

    model_id = fields.Many2one("test.enum.model")

    def create(self, values):
        if "model_id" not in values:
            values["model_id"] = self.env["test.enum.model"].create({}).id
        return super(DelegatedModel, self).create(values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf import api, fields, models
from odoo import exceptions

TEST_MODEL_NAME = "test_view_model.model"
TEST_MODEL_NAME2 = "test_view_model.model2"


class ExampleMixin(models.AbstractModel):
    _name = "example.mixin"

    is_example_mixin = True


class Model(models.Model):
    _name = TEST_MODEL_NAME

    name = fields.Char()
    value = fields.Integer()


ViewModel1 = models.ViewModel("view_model1", TEST_MODEL_NAME, mixins=["example.mixin"])


class ViewModel2(models.ViewModel("view_model2", TEST_MODEL_NAME)):
    @api.constrains("value")
    def check_value(self):
        for record in self:
            if record.value < 0:
                raise exceptions.ValidationError("value must be positive.")


class Model2(models.Model):
    _name = TEST_MODEL_NAME2

    ref_model = fields.Many2one(TEST_MODEL_NAME)
    model_value = fields.Integer(compute="get_model_value", store=True)

    @api.depends("ref_model", "ref_model.value")
    def get_model_value(self):
        for record in self:
            record.model_value = record.ref_model.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from collections import defaultdict
from inspect import getmembers
from xoeuf.odoo import api, models, tools
from xoeuf.osv.expression import Domain
from xoeuf.modules import get_caller_addon


def get_modelname(model):
    """Gets the ORM object's name for a model class/object.

    Main usage::

        self.pool[get_modelname(some_model_class)]

    :param model: Either an object (i.e an instance bound to some database) or
                  the any of the it's class definitions.

    .. deprecated:: 0.49.0 Calling with a `model proxy
                    <xoeuf.models.proxy>`:mod:.

    .. versionchanged:: 0.61.0 Calling with a `model proxy
       <xoeuf.models.proxy>`:mod: is an error.

    """
    from xoeuf.eight import string_types
    from xoeuf.odoo.models import BaseModel

    if not isinstance(model, BaseModel) and not issubclass(model, BaseModel):
        msg = "Invalid argument '%s' for param 'model'" % model
        raise TypeError(msg)
    result = model._name
    if not result:
        # This is the case of a model class having no _name defined, but then
        # it must have the _inherit and _name is regarded the same by OpenERP.
        result = model._inherit
    assert isinstance(result, string_types), "Got an invalid name for %r" % model
    return result


@property
def _onupdate_methods(self):
    """ Return a list of updater methods. """

    def is_onupdate(func):
        return callable(func) and hasattr(func, "_onupdates")

    cls = type(self)
    methods = []
    for _, func in getmembers(cls, is_onupdate):
        methods.append(func)

    # optimization: memoize result on cls, it will not be recomputed
    cls._onupdate_methods = methods
    return methods


models.BaseModel._onupdate_methods = _onupdate_methods


# extend :meth:`odoo.models.BaseModel._setup_base`
super_setup_base = models.BaseModel._setup_base


@api.model
def _setup_base(self, *args, **kwargs):
    cls = type(self)
    do_setup = not cls._setup_done
    res = super_setup_base(self, *args, **kwargs)
    if do_setup:
        cls._method_triggers = tools.Collector()
    return res


models.BaseModel._setup_base = _setup_base


# extend :meth:`odoo.models.BaseModel._setup_complete`
super_setup_complete = models.BaseModel._setup_complete


@api.model
def _setup_complete(self, *args, **kwargs):
    res = super_setup_complete(self, *args, **kwargs)
    cls = type(self)
    cls._onupdate_methods = models.BaseModel._onupdate_methods
    self.setup_triggers()
    return res


models.BaseModel._setup_complete = _setup_complete


# add :meth:`odoo.models.BaseModel.resolve_deps`
@api.model
def resolve_deps(self, method):
    """ Get the `method` dependencies list.

    Return the list of dependencies of ``method`` as tuples
    ``(model, field, path)``, where ``path`` is an optional list of field
    names.
    """
    result = []
    # add self's own dependencies
    for dotnames in method._onupdates:
        model, path = self, dotnames.split(".")
        for i, fname in enumerate(path):
            field = model._fields[fname]
            result.append((model, field, path[:i]))
            model = self.env.get(field.comodel_name)
    # add self's model dependencies
    for mname, fnames in self._depends.items():
        model = self.env[mname]
        for fname in fnames:
            field = model._fields[fname]
            result.append((model, field, None))
    # add indirect dependencies from the dependencies found above
    for model, field, path in list(result):
        for inv_field in model._field_inverses[field]:
            inv_model = self.env[inv_field.model_name]
            inv_path = None if path is None else path + [field.name]
            result.append((inv_model, inv_field, inv_path))
    return result


models.BaseModel.resolve_deps = resolve_deps


# add :meth:`odoo.models.BaseModel.setup_triggers`
@api.model
def setup_triggers(self):
    """ Add the necessary triggers to execute updater methods. """
    for update_method in self._onupdate_methods:
        for model, field, path in self.resolve_deps(update_method):
            path_str = None if path is None else (".".join(path) or "id")
            model._method_triggers.add(field, (update_method, self._name, path_str))


models.BaseModel.setup_triggers = setup_triggers


# add :meth:`odoo.models.BaseModel.update_onupdate`
@api.multi
def execute_onupdate(self, fnames):
    """Notify that fields have been modified on ``self``.

    This invalidates the cache, and prepares the recomputation of stored
    function fields (new-style fields only).

    .. versionchanged:: 0.48.0 Ignore unknown fields.

    """
    # group triggers by (model, path) to minimize the calls to search()
    triggers = defaultdict(set)
    # Take only the fields that are in the model. This is necessary because in
    # some rare cases this method is called with fields that may not be in the
    # model. (e.g. xopgi_object_merger).
    for fname in set(fnames) & set(self._fields):
        mfield = self._fields[fname]
        # group triggers by model and path to reduce the number of search()
        for method, model_name, path in self._method_triggers[mfield]:
            triggers[(model_name, method)].add(path)
    # process triggers, mark fields to be invalidated/recomputed
    for model_method, paths in triggers.items():
        model_name, method = model_method
        # process stored fields
        if paths:
            # determine records of model_name linked by any of paths to self
            if any(path == "id" for path in paths):
                target = self
            else:
                env = self.sudo().with_context({"active_test": False}).env
                target = env[model_name].search(
                    Domain.OR(*([(path, "in", self.ids)] for path in paths))
                )
                target = target.with_env(self.env)
            # prepare recomputation for each field on linked records
            method(target)


models.BaseModel.execute_onupdate = execute_onupdate


# extend :meth:`odoo.models.BaseModel._validate_fields`
super_validate_fields = models.BaseModel._validate_fields


@api.multi
def _validate_fields(self, field_names):
    res = super_validate_fields(self, field_names)
    self.execute_onupdate(field_names)
    return res


models.BaseModel._validate_fields = _validate_fields


def ViewModel(name, model_name, table=None, mixins=None):
    """Create a ViewModel in a simple way.

    A ViewModel is one that inherit other by prototype mode and is persisted
    in the same table as original model.  This allows to consider it a
    separate view of the same underlying data.  Beware you must implement all
    views consistently.

    :param name: name for view model to define.

    :param model_name: Name of the model to which the ModelView will be
                       created.

    :param table: attribute _table of model to which the ModelView will be
                  created or None to get the default using odoo convention.

    :param mixins: List of model names "mixins" that will inherit the
                   ModelView or None to include no mixin.

    """
    if not table:
        table = model_name.replace(".", "_")
    if not mixins:
        mixins = []

    class Res(models.Model):
        _name = name
        _inherit = [model_name] + mixins
        _table = table
        _module = get_caller_addon(2)  # class Res + function ViewModel.

    return Res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf import api, fields, models
from xoeuf.models.extensions import get_ref


class FooBar(models.Model):
    _name = "test_xoeuf_models.foobar"
    name = fields.Text()

    @api.model
    def get_ref(self, ref):
        return get_ref(self, ref)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf.odoo.fields import Many2one


class One2one(Many2one):
    """A classical one-2-one field.

    This is basically syntax sugar over a `~odoo.fields.Many2one`:class: with:

    - `ondelete` set to ``'cascade'`` by default;

    - a SQL uniqueness constraint in the model.

    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("ondelete", "cascade")
        super(One2one, self).__init__(*args, **kwargs)

    def setup_full(self, model):
        constraint_name = "unique_{model}_{target}".format(
            model=model._name.replace(".", "_"),
            target=self.comodel_name.replace(".", "_"),
        )
        constraints = type(model)._sql_constraints
        exists = any(True for name, _, _ in constraints if name == constraint_name)
        if not exists:
            constraints.append(
                (
                    constraint_name,
                    "unique ({field})".format(field=self.name),
                    "One2one field {field!r} must be unique".format(field=self.name),
                )
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

{
    "name": "test_api",
    "author": "Merchise Autrement [~º/~] and Contributors",
    "description": "Test for our API functions",
    "depends": ["base"],
    "installable": True,
    "auto_install": False,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# __openerp__
# ---------------------------------------------------------------------
# Copyright (c) 2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2017-08-01

{
    "name": "test_timedelta",
    "author": "Merchise Autrement [~º/~] and Contributors",
    "description": "Test for the field.Timedelta extension",
    "depends": ["base"],
    "installable": True,
    "auto_install": False,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf import fields, models

MIXIN_NAME = "example.mixin"


class ExampleMixin(models.AbstractModel):
    _name = MIXIN_NAME

    test = fields.Char(default="Hello")


class Model1(models.Model):
    _name = "test.model1"
    _inherit = MIXIN_NAME
    _description = "Model 1"


class Model2(models.Model):
    _name = "test.model2"
    _inherit = MIXIN_NAME


class SubModel2(models.Model):
    _name = "test.sub.model2"
    _inherit = "test.model2"


class TestModel(models.Model):
    _name = "test.model"

    typed_ref = fields.TypedReference(mixin=MIXIN_NAME, delegate=True)
    filtered_typed_ref = fields.TypedReference(
        mixin=MIXIN_NAME, selection=[("test.model2", "model2")]
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import contextlib
from unittest import skipIf

from hypothesis import strategies as s, given

from xoeuf import fields, MAJOR_ODOO_VERSION
from xoeuf.odoo.tests.common import TransactionCase, at_install, post_install

from ..models import COLORS, Pax, CARS, WORK_TYPE

colors = s.sampled_from(list(COLORS.__members__.values()))
color_pairs = s.sampled_from(list(COLORS.__members__.items()))
cars = s.sampled_from(list(CARS.__members__.values()))
paxs = s.sampled_from(list(Pax.__members__.values()))
wtypes = s.sampled_from(list(WORK_TYPE.__members__.values()))


@at_install(False)
@post_install(True)
class TestEnum(TransactionCase):
    def setUp(self):
        super(TestEnum, self).setUp()
        self.EnumModel = self.env["test.enum.model"]
        self.DelegatedModel = self.env["test.enum.model_delegated"]

    def test_column_type_force_char_columns(self):
        self.assertIsInstance(self.EnumModel._fields["color"], fields.Char)
        self.assertIsInstance(self.EnumModel._fields["color_name"], fields.Selection)

    def test_can_set_valid_integers(self):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"color": 1, "car": CARS.FORD, "pax": 1})
        id = obj.id
        self.EnumModel.invalidate_cache()
        obj = self.EnumModel.browse(id)
        assert obj.color == COLORS.Red, "%r == %r" % (obj.color, COLORS.Red)
        assert obj.car is CARS.FORD
        assert obj.pax == 1

    @given(colors)
    def test_can_set_valid_values(self, color):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"color": color})
        id = obj.id
        self.EnumModel.invalidate_cache()
        obj = self.EnumModel.browse(id)
        assert obj.color == color, "%r == %r" % (obj.color, color)

    @given(colors)
    def test_can_set_valid_values_as_integers_and_get_values(self, color):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"color": int(color)})
        id = obj.id
        self.EnumModel.invalidate_cache()
        obj = self.EnumModel.browse(id)
        assert obj.color == color, "%r == %r" % (obj.color, color)
        assert isinstance(obj.color, COLORS)

    def test_cannot_set_invalid_integers(self):
        # Sinces tests are run while the registry is being populated, i.e not
        # ready, we need to trick it to allow receivers be executed.
        with force_ready(self.env.registry), self.assertRaises(ValueError):
            self.EnumModel.create({"color": 10})

    def test_cannot_write_invalid_integers(self):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"color": 1})
            with self.assertRaises(ValueError):
                obj.write({"color": 10})

    def test_cannot_write_invalid_values(self):
        obj = self.EnumModel.create({"car": CARS.FORD})
        with force_ready(self.env.registry), self.assertRaises(ValueError):
            obj.write({"car": "any other brand"})

    def test_search_non_integer(self):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"car": CARS.FORD})
            self.EnumModel.invalidate_cache()
            self.assertEqual(self.EnumModel.search([("car", "=", CARS.FORD)]), obj)
            self.assertEqual(
                self.EnumModel.search([("car", "=", CARS.CHEV)], count=True), 0
            )
            self.assertEqual(
                self.EnumModel.search([("car", "in", (CARS.FORD, CARS.CHEV))]), obj
            )
            with self.assertRaises(ValueError):
                self.EnumModel.search([("car", "=", 1)])

    def test_performance(self):
        from xoeuf.fields.enumeration import EnumerationAdapter

        self.assertIn(EnumerationAdapter, type(self.EnumModel).mro())
        self.assertNotIn(EnumerationAdapter, type(self.env["res.partner"]).mro())

    def test_search_integer(self):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"color": COLORS.Red})
            self.EnumModel.invalidate_cache()
            self.assertEqual(self.EnumModel.search([("color", "=", 1)]), obj)
            with self.assertRaises(ValueError):
                self.EnumModel.search([("color", "=", 10)])

    def test_color_default_value(self):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({})
            self.assertEqual(obj.color, COLORS.Red)

    def test_color_computed_field_read(self):
        with force_ready(self.env.registry):
            obj = self.EnumModel.create({"color": COLORS.Red})
            self.assertEqual(obj.color_name, "Red")

    @given(color_pairs, color_pairs)
    def test_color_computed_field_set_on_create(self, pair, update):
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.EnumModel.create({"color_name": name})
            self.assertEqual(obj.color, value)
            name, value = update
            obj = self.EnumModel.create({"color_name": name})
            self.assertEqual(obj.color, value)
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.EnumModel.create({"color": value})
            self.assertEqual(obj.color_name, name)
            name, value = update
            obj = self.EnumModel.create({"color": value})
            self.assertEqual(obj.color_name, name)

    @given(color_pairs, color_pairs)
    def test_color_computed_field_set_on_assignment(self, pair, update):
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.EnumModel.create({})
            obj.color_name = name
            self.assertEqual(obj.color, value)
            name, value = update
            obj.color_name = name
            self.assertEqual(obj.color, value)
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.EnumModel.create({})
            obj.color = value
            self.assertEqual(obj.color_name, name)
            name, value = update
            obj.color = name
            self.assertEqual(obj.color_name, name)

    @given(color_pairs, color_pairs)
    def test_color_computed_field_set_on_write(self, pair, update):
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.EnumModel.create({})
            obj.write({"color_name": name})
            self.assertEqual(obj.color, value)
            name, value = update
            obj = self.EnumModel.create({})
            obj.write({"color_name": name})
            self.assertEqual(obj.color, value)
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.EnumModel.create({})
            obj.write({"color": value})
            self.assertEqual(obj.color_name, name)
            name, value = update
            obj = self.EnumModel.create({})
            obj.write({"color": value})
            self.assertEqual(obj.color_name, name)

    @skipIf(MAJOR_ODOO_VERSION < 12, "Supported only for  Odoo 12+")
    @given(color_pairs, color_pairs)
    def test_color_computed_field_set_on_create_delegated(self, pair, update):
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.DelegatedModel.create({"color_name": name})
            self.assertEqual(obj.color, value)
            name, value = update
            obj = self.DelegatedModel.create({"color_name": name})
            self.assertEqual(obj.color, value)
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.DelegatedModel.create({"color": value})
            self.assertEqual(obj.color_name, name)
            name, value = update
            obj = self.DelegatedModel.create({"color": value})
            self.assertEqual(obj.color_name, name)

    @skipIf(MAJOR_ODOO_VERSION < 12, "Supported only for  Odoo 12+")
    @given(color_pairs, color_pairs)
    def test_color_computed_field_set_on_assignment_delegated(self, pair, update):
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.DelegatedModel.create({})
            obj.color_name = name
            self.assertEqual(obj.color, value)
            name, value = update
            obj.color_name = name
            self.assertEqual(obj.color, value)
        with force_ready(self.env.registry):
            name, value = pair
            obj = self.DelegatedModel.create({})
            obj.color = value
            self.assertEqual(obj.color_name, name)
            name, value = update
            obj.color = name
            self.assertEqual(obj.color_name, name)


@contextlib.contextmanager
def force_ready(registry):
    ready = registry.ready
    registry.ready = True
    try:
        yield
    finally:
        registry.ready = ready



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Xœuf services for access Open Object (OpenERP) models.

Implement the network protocols that the Xœuf applications that access OpenERP
databases uses to communicate with its clients.

"""

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import pytz
from xoeuf.tools import dt_as_timezone


def savepoint(cr, name=None):
    """A context manager that enters a new savepoint.

    If `name` is not provided a random one is generated.

    .. note:: In Odoo v8, the cursor `cr` has already a ``savepoint`` method.
              This function will fallback to it in this case.  However, the
              argument `name` will be ignored in this case.

    """
    from xoutil.symbols import Unset

    _savepoint = getattr(cr, "savepoint", Unset)
    if _savepoint is Unset:
        from contextlib import contextmanager

        if not name:
            from xoutil.uuid import uuid

            name = uuid(True)

        @contextmanager
        def _savepoint():
            cr.execute('SAVEPOINT "%s"' % name)
            try:
                yield
            except:  # noqa
                cr.execute('ROLLBACK TO SAVEPOINT "%s"' % name)
                raise
            else:
                cr.execute('RELEASE SAVEPOINT "%s"' % name)

    return _savepoint()


def datetime_user_to_server_tz(cr, uid, userdate, tz_name=None, context=None):
    """ Convert date values expressed in user's timezone to
    server-side UTC timestamp.

    :param datetime userdate: datetime in user time zone
    :return: UTC datetime for server-side use
    """
    from xoeuf import api

    utc = pytz.UTC
    if userdate.tzinfo:
        return utc.normalize(userdate)
    if not tz_name:
        env = api.Environment(cr, uid, context or {})
        user = env.user
        if user.tz:
            dt = dt_as_timezone(userdate, user.tz)
        else:
            dt = dt_as_timezone(userdate)
    else:
        dt = dt_as_timezone(userdate, tz_name)
    return utc.normalize(dt)


def datetime_server_to_user_tz(cr, uid, serverdate, tz_name=None, context=None):
    """ Convert date values expressed in server-side UTC timestamp to
    user's timezone.

    :param datetime serverdate: datetime in server-side UTC timestamp.
    :return: datetime on user's timezone
    """
    from xoeuf import api

    dt = dt_as_timezone(serverdate)  # datetime in UTC
    if not tz_name:
        env = api.Environment(cr, uid, context or {})
        user = env.user
        user_tz = pytz.timezone(user.tz) if user.tz else pytz.UTC
    else:
        user_tz = pytz.timezone(tz_name)
    return user_tz.normalize(dt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf import fields
from xoeuf.odoo.tests.common import TransactionCase


class TestReferences(TransactionCase):
    def test_simple_typed_reference(self):
        obj = self.env["test.model"].create({})
        typed_ref = obj._fields["typed_ref"]
        self.assertTrue(isinstance(typed_ref, fields.TypedReference))
        self.assertEqual(typed_ref.mixin, "example.mixin")

        def typed_ref_assign(value):
            obj.typed_ref = value
            return True

        self.assertTrue(typed_ref_assign(False))
        obj2 = self.env["test.model1"].create({})
        self.assertTrue(typed_ref_assign(obj2))
        self.assertTrue(typed_ref_assign("%s,%d" % (obj2._name, obj2.id)))
        self.assertRaises(ValueError, typed_ref_assign, obj)

    def test_filtered_typed_reference(self):
        obj = self.env["test.model"].create({})

        def typed_ref_assign(value):
            obj.filtered_typed_ref = value
            return True

        obj2 = self.env["test.model1"].create({})
        # This should work, but the odoo validation is missing,
        # I am making a PR to propose it.
        # self.assertRaises(ValueError, typed_ref_assign, '%s,%d' % (obj2._name, obj2.id))
        self.assertRaises(ValueError, typed_ref_assign, obj2)

    def test_delegate_typed_reference(self):
        obj = self.env["test.model"].create({})
        obj2 = self.env["test.model1"].create({"test": "ok"})
        obj.typed_ref = obj2
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertTrue(obj._fields["typed_ref"].delegate)
        self.assertTrue(obj._fields["test"].compute)
        self.assertEqual(obj.test, obj2.test)
        obj.test = "other value"
        self.assertEqual(obj.test, obj2.test)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.test, obj2.test)
        obj2.test = "other more value"
        # todo: indirect depends are not working fine.
        # self.assertEqual(obj.test, obj2.test)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.test, obj2.test)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Some tools to improve `OpenERP` security.

- :func:`reset_all_passwords`: to reset all passwords in a data-base.

- :func:`reset_invalid_passwords`: to reset all invalid passwords in a
  data-base.

Previous two functions uses `xoutil.crypto.generate_password` to generate new
passwords using as `pass_phrase` the user login, `level` means a generation
method.  Each level implies all other with an inferior numerical value.  See
`xoutil.crypto.generate_password` for more information about defined
constants of security level.

"""

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)


from xoutil.crypto import (  # noqa
    PASS_PHRASE_LEVEL_BASIC,
    PASS_PHRASE_LEVEL_MAPPED,
    PASS_PHRASE_LEVEL_MAPPED_MIXED,
    PASS_PHRASE_LEVEL_MAPPED_DATED,
    PASS_PHRASE_LEVEL_STRICT,
    DEFAULT_PASS_PHRASE_LEVEL as _DEF_LEVEL,
)


from xoutil.names import strlist as strs

__all__ = strs(
    "PASS_PHRASE_LEVEL_BASIC",
    "PASS_PHRASE_LEVEL_MAPPED",
    "PASS_PHRASE_LEVEL_MAPPED_MIXED",
    "PASS_PHRASE_LEVEL_MAPPED_DATED",
    "PASS_PHRASE_LEVEL_STRICT",
    "reset_all_passwords",
    "reset_invalid_passwords",
)
del strs


def _reset_passwords(self, security_level, verbose, check=None):
    """Internal module function to reset passwords in a data-base.

    This function is used by :func:`reset_all_passwords` and
    :func:`reset_invalid_passwords` functions.

      :param self: The res.user model.

      :param security_level: Numerical security level (the bigger the more
             secure).

      :param verbose: If True, print to ``stdout`` information of every
             password change.

      :param check: A function that checks if for a given user the password
             must be changed or not (if not given, is equivalent for True to
             all users).  It must has following definition::

               def check(self, cr, id, login):

             Where `self` is the ``res.users`` model, `cr` the active
             data-base cursor, `id` the user data-base id to check, and
             `login` the user login identifier.

    """
    from xoutil.crypto import generate_password
    from xoutil.future.codecs import safe_encode

    users = self.search([])
    for user in users:
        login = user.login
        if check is None or check(user):
            user.password = password = generate_password(login, security_level)
            if verbose:
                print(
                    safe_encode(
                        ">>> id: %(id)s, login: %(login)s, "
                        "name: %(name)s, "
                        "password: '%(password)s'"
                        % dict(
                            id=user.id,
                            login=user.login,
                            name=user.name,
                            password=password,
                        )
                    )
                )


def reset_all_passwords(self, security_level=_DEF_LEVEL, verbose=True):
    """Reset all passwords in a data-base.

    :param self: The res.users model.

    :param security_level: Numerical security level (the bigger the more
           secure).

    :param verbose: If True, print every change password to ``stdout``.

    This function can be used as::

      from xoeuf.security import reset_all_passwords
      reset_all_passwords(env['res.users'], security_level=2)

    See module documentation for more info.

    """
    _reset_passwords(self, security_level, verbose)


def reset_invalid_passwords(self, security_level=_DEF_LEVEL):
    """Reset all invalid passwords in a data-base.

    :param self: The res.users model.

    :param security_level: Numerical security level (the bigger the more
           secure).

    An invalid password is when it is the same as login name. Print
    information about all users with invalid passwords.

    This function can be used as::

      from xoeuf.security import reset_invalid_passwords
      reset_invalid_passwords(env['res.users'])

    See module documentation for more info.

    """

    def check(self):
        from xoeuf.odoo.exceptions import AccessDenied

        try:
            self.sudo(self.id).check_credentials(self.login)
            return True
        except AccessDenied:
            return False

    _reset_passwords(self, security_level, True, check)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Implements the meta() function using the Py3k syntax.

"""
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_imports,
)

from . import _py3

assert _py3, "This module should be loaded in Py3k only"


def metaclass(meta, **kwargs):
    from ._meta import Mixin

    if isinstance(meta, type) and issubclass(meta, type) and meta is not type:
        metabase = meta.__base__
    else:
        metabase = type

    class inner_meta(metabase):
        @classmethod
        def __prepare__(cls, name, bases, **kwargs):
            real = name != "__inner__"
            prepare = getattr(meta, "__prepare__", None) if real else None
            return prepare(name, bases, **kwargs) if prepare else dict()

        def __new__(cls, name, bases, attrs, **kw):
            if name != "__inner__":
                bases = tuple(b for b in bases if Mixin not in b.__bases__)
                return meta(name, bases, attrs)
            else:
                return type.__new__(cls, name, bases, attrs)

        def __init__(self, name, bases, attrs, **kw):
            pass

    from ._types import new_class

    kwds = dict(kwargs, metaclass=inner_meta)
    return new_class("__inner__", (Mixin,), kwds=kwds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Implements the metaclass() function using the Py3k syntax.

"""
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_imports,
)

from . import _py3

assert not _py3, "This module should not be loaded in Py3k"


METACLASS_ATTR = "__metaclass__"


def metaclass(meta, **kwargs):
    from ._meta import Mixin

    prepare = getattr(meta, "__prepare__", None)
    if prepare:
        import warnings

        warnings.warn(
            "Python 2.7 does not have the __prepare__ stuff and "
            'the metaclass "%s" seems to needs it.' % meta,
            stacklevel=2,
        )

    if isinstance(meta, type) and issubclass(meta, type) and meta is not type:
        metabase = meta.__base__
    else:
        metabase = type

    class inner_meta(metabase):
        def __new__(cls, name, bases, attrs, **kw):
            from copy import copy

            if name != "__inner__":
                bases = tuple(b for b in bases if Mixin not in b.__bases__)
                from ._types import prepare_class

                kwds = dict(kwargs, metaclass=meta)
                basemeta, _ns, kwds = prepare_class(name, bases, kwds=kwds)
                ns = copy(_ns)
                update = getattr(ns, "update", None)
                if update:
                    update(attrs)
                else:
                    for attr, val in attrs.items():
                        ns[attr] = val
                if METACLASS_ATTR not in ns:
                    ns[METACLASS_ATTR] = meta
                return basemeta(name, bases, ns)
            else:
                return type.__new__(cls, name, bases, attrs)

        def __init__(self, name, bases, attrs, **kw):
            pass

    from ._types import new_class

    kwds = dict(kwargs, metaclass=inner_meta)

    def exec_body(ns):
        return ns

    return new_class("__inner__", (Mixin,), kwds=kwds, exec_body=exec_body)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""(Old fashioned) extensions to fields in the ORM.

"""

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)


from xoutil.deprecation import deprecated

# FIX: Next doesn't work any more in version 10
from openerp.osv import fields as _v7_fields

import pytz
from xoeuf.tools import localtime_as_remotetime
import xoeuf.fields as _future


try:
    integer_types = (int, long)
except NameError:
    integer_types = (int,)


@deprecated(_future.LocalizedDatetime)
class localized_datetime(_v7_fields.function):
    """A field for localized datetimes.

    Localized datetimes are actually a functional field that takes two
    underlying columns a datetime and timezone name.

    :param dt: The name of the column where the datetime (in UTC) is actually
               saved.

    :param tzone:  The name of the column where the timezone where the event
                   should be presented to the user.

    Upon reading, we assume the user's timezone is properly set.

    The datetime column will be actually saved in UTC as all datetimes in
    Odoo.   But upon reading we convert to a properly shifted datetime so that
    is presented to the user in the saved time zone.

    .. note:: At the time this field is read-only, not searchable, and
              non-storable.

    """

    def _ldt_write(self, obj, cr, uid, ids, field, val, args, context=None):
        tzone_field = self.__tzone_field
        dt_field = self.__dt_field
        tz = context.get("tz", None)
        if not tz:
            user = obj.pool["res.users"].browse(cr, uid, uid, context=context)
            tz = pytz.timezone(user.tz) if user.tz else pytz.UTC
        else:
            tz = pytz.timezone(tz)
        if isinstance(ids, integer_types):
            ids = [ids]
        for row in obj.browse(cr, uid, ids, context=context):
            tzone = getattr(row, tzone_field)
            if not tzone:
                tzone = pytz.UTC
            else:
                tzone = pytz.timezone(tzone)
            # Compute the datetime in users timezone, then force to it the
            # desired TZ and back to UTC.
            if val:
                val = localtime_as_remotetime(val, tz, tzone)
            obj.write(cr, uid, [row.id], {dt_field: val}, context=context)

    def _ldt_read(self, obj, cr, uid, ids, field, arg, context=None):
        tzone_field = self.__tzone_field
        dt_field = self.__dt_field
        tz = context.get("tz", None)
        if not tz:
            user = obj.pool["res.users"].browse(cr, uid, uid, context=context)
            tz = pytz.timezone(user.tz) if user.tz else pytz.UTC
        else:
            tz = pytz.timezone(tz)
        if isinstance(ids, integer_types):
            ids = [ids]
        res = {}
        for row in obj.browse(cr, uid, ids, context=context):
            tzone = getattr(row, tzone_field)
            if not tzone:
                tzone = pytz.UTC
            else:
                tzone = pytz.timezone(tzone)
            dt = getattr(row, dt_field)
            # Compute the datetime in the desired timezone, then extract
            # all datetime components but the TZ and localize it to the
            # users TZ and convert it back to UTC... This makes the UI to
            # reverse the process and show the datetime in the desired
            # timezone.
            if dt:
                dt = localtime_as_remotetime(dt, tzone, tz)
            res[row.id] = dt
        return res

    def __init__(self, dt_field, tzone_field, **kwargs):
        self.__dt_field = dt_field
        self.__tzone_field = tzone_field
        super(localized_datetime, self).__init__(
            self._ldt_read, None, fnct_inv=self._ldt_write, store=False, **kwargs
        )


del _future, deprecated



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf.odoo.tests.common import TransactionCase
from xoeuf.odoo.exceptions import ValidationError


class TestViewModel(TransactionCase):
    def test_view_model_module(self):
        self.assertEqual(self.env["view_model1"]._module, "test_view_model")
        self.assertEqual(self.env["view_model2"]._module, "test_view_model")

    def test_mixins(self):
        is_example_mixin = lambda model_name: getattr(
            self.env[model_name], "is_example_mixin", False
        )
        self.assertTrue(is_example_mixin("view_model1"))
        self.assertFalse(is_example_mixin("view_model2"))
        self.assertFalse(is_example_mixin("test_view_model.model"))

    def test_non_shared_extensions(self):
        view1 = self.env["view_model1"]
        view2 = self.env["view_model2"]
        values = {"name": "x", "value": -10}
        self.assertTrue(view1.create(values))
        self.assertRaises(ValidationError, view2.create, values)

    def test_depends(self):
        model1 = self.env["test_view_model.model"]
        model2 = self.env["test_view_model.model2"]
        view2 = self.env["view_model1"]
        values = {"name": "x", "value": -10}
        obj1 = model1.create(values)
        obj2 = model2.create({"ref_model": obj1.id})
        model1.invalidate_cache()
        model2.invalidate_cache()
        self.assertEqual(obj1.value, obj2.model_value)
        obj1.value = 3
        model1.invalidate_cache()
        model2.invalidate_cache()
        self.assertEqual(obj1.value, obj2.model_value)
        view2.browse(obj1.id).value = 7
        model1.invalidate_cache()
        model2.invalidate_cache()
        self.assertNotEqual(obj1.value, obj2.model_value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from datetime import datetime, time
from functools import partial
from xoeuf.eight import string_types

from odoo.fields import Selection, Datetime, Float, Default as DEFAULT

from xoeuf.tools import normalize_datetime, get_time_from_float

from .utils import TimeRange as _TimeRangeObject


class TimeField(_TimeRangeObject):
    def __init__(self, start, end, name):
        super(TimeField, self).__init__(start, end)
        self.name = name


class TimeRangeSelector(object):
    def __init__(self, choices=()):
        self.ranges = [TimeField(c[2], c[3], c[0]) for c in choices]

    def get_range(self, _time=time.min):
        if isinstance(_time, datetime):
            _time = _time.time()
        return next((_range for _range in self.ranges if _time in _range), None)


# TODO: This is not actually a Selection.  In a selection the user is allowed
# to select a value from a collection.  This is kind of computation of a set
# from a given classifier.
class TimeRange(Selection):
    """A timerange field.

    .. warning:: This field is still experimental and may heavily changed or
       removed.

    A field calculated from two elements, the field that contains the time of day
    represented either by a 'datetime' or a 'float' and the list of possible
    values of time ranges in a day.

    :param t_field: The name of the column that contains the day time.

    :param selection: The possible ranges for this field.  It is given as
        either a list of tuples ``(value, string, start, end)``, a model
        method, or a method name.  `value` is the range identifier; `string`
        is the translatable text to show; `start` and `end` are values of
        times expressed in the 'hour:minutes' format.

    :param selection_add: Adds more options to the `selection`.  Only needed
                          if your extending an existing model's field.

    """

    type = "timerange"

    _slots = {"time_field": None, "readonly": True}

    def __init__(self, time_field, selection=DEFAULT, *args, **kwargs):
        from xoutil.symbols import Unset

        kwargs = dict(dict(copy=False, readonly=True), **kwargs)
        super(TimeRange, self).__init__(
            time_field=time_field or Unset, selection=selection, *args, **kwargs
        )
        self.time_field = time_field or Unset

    def new(self, **kwargs):
        # Pass original args to the new one.  This ensures that the
        # t_field and selection are present.  In odoo/models.py, Odoo calls
        # this `new()` without arguments to duplicate the fields from parent
        # classes.
        return type(self)(self.time_field, **kwargs)

    def setup_full(self, model):
        super(TimeRange, self).setup_full(model)
        time_field = self.time_field
        field = model._fields[time_field]
        if not isinstance(field, (Datetime, Float)):
            raise TypeError(
                "Type of time_field must be Datetime or Float, "
                "instead of %s." % type(field)
            )

    def _setup_regular_full(self, env):
        # This is to support the case where ModelB `_inherits` from a ModelA
        # with a timerange. In such a case, we don't override the
        # compute method.
        super(TimeRange, self)._setup_regular_full(env)
        self.depends = (self.time_field,)
        self.compute = self._compute

    def _description_selection(self, env):
        """ return the selection list (tuple (value, label, start, end)); labels
            are translated according to context language
        """
        selection = self.selection
        if isinstance(selection, string_types):
            return getattr(env[self.model_name], selection)()
        if callable(selection):
            return selection(env[self.model_name])

        # translate selection labels
        if env.lang:
            name = "%s,%s" % (self.model_name, self.name)
            translate = partial(
                env["ir.translation"]._get_source, name, "selection", env.lang
            )
            return [
                (value, translate(label), start, end if label else label)
                for value, label, start, end in selection
            ]
        else:
            return selection

    def _dict_selection(self, env):
        """ return a dictionary value:translatable String """
        selection = self.selection
        if env.lang and isinstance(selection, list):
            name = "%s,%s" % (self.model_name, self.name)
            translate = partial(
                env["ir.translation"]._get_source, name, "selection", env.lang
            )
            return {
                value: translate(label) if label else label
                for value, label, start, end in selection
            }
        if isinstance(selection, string_types):
            selection = getattr(env[self.model_name], selection)()
        if callable(selection):
            selection = selection(env[self.model_name])
        return {value: label for value, label, start, end in selection}

    def get_values(self, env):
        """ return a list of the possible values """
        selection = self.selection
        if isinstance(selection, string_types):
            selection = getattr(env[self.model_name], selection)()
        elif callable(selection):
            selection = selection(env[self.model_name])
        return [value for value, _, _, _ in selection]

    def _compute(self, records):
        time_field = self.time_field
        env = records.env
        field = records._fields[time_field]
        for item in records:
            t_value = getattr(item, time_field)
            if isinstance(field, Datetime):
                dt = normalize_datetime(t_value)
                _time = dt.time()
            if isinstance(field, Float):
                _time = get_time_from_float(t_value)
            setattr(item, self.name, self._compute_selection(_time, env))

    def _compute_selection(self, time_value, env):
        selections = self._description_selection(env)
        _range = TimeRangeSelector(selections).get_range(time_value)
        if _range:
            return _range.name
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Tools for "Query Language".

Most of this has been extracted from `xotl.ql` (but we must support Python
2.7).

"""
from __future__ import division, print_function, absolute_import

import ast as pyast
from xoutil.future.types import new_class

import sys

_py_version = sys.version_info
_py3 = _py_version >= (3, 0)
del sys


# For each AST class in the Python ast module we build a new one here that
# supports equality comparison::
#
#    qst.Name('a', qst.Load()) == qst.Name('b', qst.Load())
#
class PyASTNode(object):
    def __eq__(self, other):
        from operator import eq

        Unset = object()  # xoutil's Unset is equal to 0!
        res = True
        i = 0
        # Explicitly deal with NameConstant None.
        if is_constant(self, None):
            return other is None or is_constant(other, None) or other == LOAD_NONE
        elif is_constant(other, None):
            return self is None or self == LOAD_NONE
        attrs = self._fields
        if not attrs:
            # If no attrs only check for typing, both should have the same
            # name (_ast.Load and qst.Load, etc.) and one must a subclass of
            # the other.
            self, other = type(self), type(other)
            attrs = ("__name__",)
            res = issubclass(self, other) or issubclass(other, self)
        get_from_source = lambda a: getattr(self, a, Unset)
        get_from_target = lambda a: getattr(other, a, Unset)
        while res and (i < len(attrs)):
            attr = attrs[i]
            sattr = get_from_source(attr)
            tattr = get_from_target(attr)
            if eq(sattr, tattr):
                i += 1
            elif sattr is Unset or tattr is Unset:
                res = False
            elif sattr is None and tattr == LOAD_NONE or tattr == NONE_CT:
                i += 1
            elif tattr is None and sattr == LOAD_NONE or sattr == NONE_CT:
                i += 1
            else:
                res = False
        return res

    __hash__ = None

    def __str__(self):
        def r(who):
            if isinstance(who, pyast.AST):
                return "<ast: %s>" % type(who).__name__
            else:
                return repr(who)

        res = []
        children = [(self, None, 0)]
        while children:
            child, field, depth = children.pop(0)
            if field:
                res.append(" " * 3 * depth + "{}: ".format(field) + r(child))
            else:
                res.append(" " * 3 * depth + r(child))
            fields = getattr(child, "_fields", [])
            grandchildren = [(getattr(child, f), f, depth + 1) for f in fields]
            if grandchildren:
                # If any grandchild is a list 'expand it', this helps to get a
                # nicer visualization.
                i = 0
                while i < len(grandchildren):
                    val, f, d = grandchildren[i]
                    if isinstance(val, (list, tuple)):
                        j = len(val)
                        grandchildren[i : i + 1] = [
                            (v, f + "[%d]" % k, d) for k, v in enumerate(val)
                        ]
                        i += j
                    else:
                        i += 1
                children[0:0] = grandchildren
        return "\n".join(res)

    @classmethod
    def from_pyast(cls, node):
        """Convert an `ast.AST`:py:class: node into the equivalent qst node."""

        def _convert(value):
            if value:
                type_ = globals().get(type(value).__name__, None)
                if type_:
                    value = type_.from_pyast(value)
            return value

        assert cls.__name__ == type(node).__name__
        res = cls()
        for attr in node._attributes:
            setattr(res, attr, getattr(node, attr, None))
        for field in node._fields:
            value = getattr(node, field, None)
            if isinstance(value, list):
                value = [_convert(val) for val in value]
            elif isinstance(value, tuple):
                value = tuple(_convert(val) for val in value)
            else:
                value = _convert(value)
            setattr(res, field, value)
        return res


__all__ = []
_nodes = [
    pyast.Expression,
    pyast.expr,
    pyast.boolop,
    pyast.unaryop,
    pyast.keyword,
    pyast.slice,
    pyast.operator,
    pyast.cmpop,
    pyast.comprehension,
    pyast.arguments,
    pyast.expr_context,
]

if _py3:
    _nodes.append(pyast.arg)

_current = 0
while _current < len(_nodes):
    _node = _nodes[_current]
    _more = _node.__subclasses__()  # Don't place this after the new class.

    globals()["_PyAst_%s" % _node.__name__] = _node
    # Has a constructor create the class for comparing
    globals()[_node.__name__] = cls = new_class(
        _node.__name__, bases=(_node, PyASTNode)
    )
    cls.__module__ = "qst"
    __all__.append(_node.__name__)

    if _more:
        _nodes.extend(_more)
    _current += 1

if (3, 6) <= _py_version:
    # In Python 3.6+:
    #
    #      comprehension = (expr target, expr iter, expr* ifs, int is_async)
    #
    # I don't think this async stuff will impact the byte-code of a normal
    # query-like comprehension.
    #
    class comprehension(comprehension):  # noqa
        def __init__(self, *args):
            if 0 < len(args) < 4:
                args += (False,)
            super().__init__(*args)


# This None as a name.  Only use this for comparison, not as a return value.
LOAD_NONE = Name("None", Load())  # noqa

try:
    NONE_CT = NameConstant(None)  # noqa
except NameError:

    class NameConstant(object):
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            return is_constant(other, self.value)

    NONE_CT = NameConstant(None)


def is_constant(which, value):
    "Test if which is a NameConstant for `value`."
    return isinstance(which, NameConstant) and which.value is value  # noqa


def parse(source, filename="<unknown>", mode="eval"):
    assert mode == "eval"
    res = pyast.parse(source, filename, mode)
    return Expression.from_pyast(res)  # noqa


def ensure_compilable(st):
    visitor = SetAttributesVisitor(lineno=1, col_offset=0)
    visitor.visit(st)
    return st


class SetAttributesVisitor(pyast.NodeVisitor):
    def __init__(self, **attrs):
        self.attrs = attrs

    def generic_visit(self, node):
        from xoutil.symbols import Unset

        get = lambda a: getattr(node, a, Unset)
        for attr, val in self.attrs.items():
            if get(attr) is Unset:
                setattr(node, attr, val)
        return super(SetAttributesVisitor, self).generic_visit(node)


def make_argless_call(fn):
    return make_call(fn)


def make_call(fn, *args, **kwargs):
    kws = [keyword(name, val) for name, val in kwargs.items()]  # noqa
    if _py3:
        return Call(fn, list(args), kws)  # noqa
    else:
        return Call(fn, list(args), kws, None, None)  # noqa


def make_arguments(*names):
    if _py3:
        if _py_version >= (3, 4):
            return arguments(  # noqa
                [arg(name, None) for name in names], None, [], [], None, []  # noqa
            )
        else:
            return arguments(  # noqa
                [arg(name, None) for name in names],  # noqa
                None,
                None,
                [],
                None,
                None,
                [],
                [],
            )
    else:
        return arguments(  # noqa
            [Name(name, Param()) for name in names], None, None, []  # noqa
        )


def make_attr(node, attr):
    return Attribute(node, attr, Load())  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)


from xoeuf import models, fields


class DomainModel(models.Model):
    _name = "test_domain.model"

    name = fields.Text()
    age = fields.Integer()

    parent_id = fields.Many2one(_name)
    children_ids = fields.One2many(_name, "parent_id")

    def __repr__(self):
        def _repr(rec):
            return "{id}({name!r}, {age!r}, {children!r})".format(
                id=rec.id, name=rec.name, age=rec.age, children=rec.children_ids
            )

        recs = "; ".join(_repr(r) for r in self)
        return "{{{0}}}".format(recs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from xoeuf.odoo.tests.common import TransactionCase


class TestViewModel(TransactionCase):
    def test_shared_extensions(self):
        self.assertTrue(
            all(
                "new_value" in self.env[model]._fields
                for model in ["test_view_model.model", "view_model1", "view_model2"]
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Implements the metaclass() function using the Py3k syntax.

"""
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from . import _py3

if _py3:
    from ._meta3 import metaclass
else:
    from ._meta2 import metaclass


metaclass.__module__ = __name__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import argparse
import io
import os
import pprint
import textwrap

from xoeuf.odoo import SUPERUSER_ID
from xoeuf.odoo.fields import date, Date
from xoeuf.odoo.modules.module import get_module_path

from . import Command

MIGRATION_TEMPLATE = """#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# {migration_name}
# ---------------------------------------------------------------------
# Copyright (c) {current_year} Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on {today}

'''{migration_readable_name}

'''


CURRENT_MODULE = '{current_module}'
SOURCE_MODULE = '{source_module}'
FIELDS = {fields}

def migrate(cr, version):
    if version:
        cr.execute(
            '''
            UPDATE ir_model_data
            SET module=%(CURRENT_MODULE)s
            WHERE module=%(SOURCE_MODULE)s AND model='ir.model.fields' AND
            name NOT IN (SELECT name
                         FROM ir_model_data
                         WHERE module=%(CURRENT_MODULE)s) AND
            name IN %(FIELDS)s
            ''',
            {{
                'CURRENT_MODULE': CURRENT_MODULE,
                'SOURCE_MODULE': SOURCE_MODULE,
                'FIELDS': FIELDS
            }}
        )
        cr.commit()
"""

INPUT_VALID_VALUES = [
    (True, ("y", "ye", "yes", "s", "si", "sí")),
    (False, ("n", "no", "")),
]


def newline():
    print("")


def write(msg):
    print(msg)


def read_bool_value(prompt=None):
    from xoeuf.eight import input

    _input = input(prompt)
    _input = _input.lower()
    for result, valid_values in INPUT_VALID_VALUES:
        if _input in valid_values:
            return result
    raise ValueError("Invalid option.")


def make_reference_name(field):
    """Get ir_model_data name generated for given model-field pair.

    """
    return "field_%s_%s" % (field[0].replace(".", "_"), field[1])


def get_fields(db, src):
    """ Get all fields created on `src` module.

    """
    with db(transactional=True) as cr:
        cr.execute(
            """
            SELECT model, name
            FROM ir_model_fields
            WHERE id IN (
                SELECT res_id
                FROM ir_model_data
                WHERE module=%s AND model='ir.model.fields')
            ORDER BY model, name
        """,
            (src.module_name,),
        )
        for field in cr.fetchall():
            yield (make_reference_name(field), field)


def _get_content(src, dst, file_name, fields):
    fields_pretty_print = io.BytesIO()
    pprint.pprint(tuple(fields), stream=fields_pretty_print)
    name = textwrap.fill(
        "Migrate (some) fields reference from %s to %s"
        % (src.module_name, src.module_name),
        78,
    )
    today = date.today()
    return MIGRATION_TEMPLATE.format(
        migration_name=file_name,
        current_year=today.year,
        today=Date.to_string(today),
        migration_readable_name=name,
        source_module=src.module_name,
        current_module=dst.module_name,
        fields=fields_pretty_print.getvalue(),
    )


def write_migration(file_path, file_name, content):
    if not os.path.exists(file_path):
        os.makedirs(file_path)
    with open(file_path + file_name, "wb") as f:
        f.write(content)


def check_fields(fields, valid_fields):
    if fields[0].lower() in ("a", "all"):
        result = [ref for ref, _ in valid_fields]
    else:
        result = []
        for field in fields:
            field = field.strip().rsplit(".", 1)
            if field and len(field) == 2:
                ref = make_reference_name(field)
                result.append(ref)
                if (ref, tuple(field)) not in valid_fields:
                    raise ValueError("Given field (%s, %s) is not valid." % field[-1])
    return result


class Migration(Command):
    """ Generates a migration file to move field references from a module to
    an other.

    """

    @classmethod
    def _get_arg_parser(cls):
        res = getattr(cls, "_arg_parser", None)
        if not res:
            res = argparse.ArgumentParser()
            cls._arg_parser = res
            res.add_argument(
                "-d",
                "--database",
                dest="database",
                required=True,
                type=DB,
                help="Name of the database used used to check actual status.",
            )
            res.add_argument(
                "src_module", type=Module, help="Name of the source module"
            )
            res.add_argument(
                "dst_module", type=Module, help="Name of the destination module"
            )
            res.add_argument(
                "-f",
                "--fields",
                dest="fields",
                nargs="+",
                help="Field to migrate on comma separated model.field "
                "format or `All` for all fields.",
            )
        return res

    def run(self, args):
        parser = self._get_arg_parser()
        options = parser.parse_args(args)
        options.src_module.load_module(options.database)
        options.dst_module.load_module(options.database)
        if options.fields:
            fields = check_fields(
                options.fields,
                list(get_fields(options.database.db, options.src_module)),
            )
        else:
            fields = []
            newline()
            write(
                "Press y/n to select what fields include on migration. "
                "Yes response must be explicit (<Enter> means no)."
            )
            for ref, model_field in get_fields(options.database.db, options.src_module):
                if read_bool_value("Model: %s; Field: %s >>> " % model_field):
                    fields.append(ref)
        if not fields:
            raise ValueError("No field selected")
        file_name = options.dst_module.get_file_name(options.src_module)
        content = _get_content(
            options.src_module, options.dst_module, file_name, fields
        )
        try:
            write_migration(options.dst_module.get_migration_path(), file_name, content)
        except Exception as e:
            write(str(e))
            newline()
            write(
                "\nCopy following lines and create manually "
                "the corresponding migration with this content."
            )
            write(content)
        else:
            write("Migration is created successfully.")


class DB(object):
    def __init__(self, db_name):
        import importlib

        py_module = "xoeuf.pool.%s" % db_name
        self.db = importlib.import_module(py_module)


class Module(object):
    def __init__(self, module_name):
        self.module_name = module_name
        self.module_path = get_module_path(module_name, display_warning=False)
        self.new_version = None
        self.module = None

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__.get(item)
        else:
            return getattr(self.module, item)

    def get_migration_path(self):
        return "%s/migrations/%s/" % (self.module_path, self.new_version)

    def get_file_name(self, src):
        """ Get a valid file name for the migration module.

        """
        base_name = "pre-migrate_fields_reference_from_%s_to_%s" % (
            src.module_name,
            self.module_name,
        )
        name = base_name
        count = 0
        while os.path.exists(self.get_migration_path() + name + ".py"):
            count += 1
            name = base_name + str(count)
        return name + ".py"

    def load_module(self, db):
        with db.db(transactional=True) as cr:
            obj = db.db.models.ir_module_module
            m_id = obj.search(
                cr, SUPERUSER_ID, [("name", "=", self.module_name)], limit=1
            )
            if m_id:
                module = obj.browse(cr, SUPERUSER_ID, m_id)
            if not module:
                raise ValueError("Not found module %s." % self.module_name)
            if module.state == "uninstalled":
                raise ValueError(
                    "Module %s is not already installed." % self.module_name
                )
            write(module.state + module.latest_version)
            self.module = module
            self.get_new_version()

    def get_new_version(self):
        if self.latest_version:
            current_version = [v for v in self.latest_version.split(".")]
            try:
                minor_version = int(current_version[-1])
                current_version[-1] = str(minor_version + 1)
            except ValueError:
                current_version.append("1")
            self.new_version = ".".join(current_version)


def main():
    from xoutil.cli.app import main
    from xoutil.cli import command_name

    main(default=command_name(Migration))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

from . import test_localized_dt  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""An example of an application that use :mod:`xoeuf.cli`.

It behaves similar to "openerp-server" script. This module does not provide any
external facilities, but uses :func:`xoutil:xoutil.cli.app.main` to run the
OpenERP server. Usage::

  $ python server.py [options...]

"""

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)


def server():
    from xoeuf.cli import DEFAULT_COMMAND
    from xoutil.cli.app import main

    main(default=DEFAULT_COMMAND)


if __name__ == "__main__":
    server()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Extensions to `odoo.osv.expression`.

This module reexport all symbols from the `odoo.osv.expression` so it can be
used a replacement.

Additions and changes:

- There's a new `Domain`:class: which allows to do some arithmetic with
  domains.

- The functions `AND`:func: and `OR`:func: don't need to you pass a domain in
  first normal form, they ensure that themselves.

- You can test some a weak form of implication, i.e 'Domain(X).implies(Y)' is
  True if can proof that whenever X is True, Y also is.  This method can have
  false negatives, but not false positives: there are cases for which we can't
  find the proof (we return False) which should be True; but if we return
  True, there's a proof.

"""
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import operator
from itertools import chain
from xoeuf.odoo.osv import expression as _odoo_expression
from xoeuf.eight import string_types

from . import ql


# TODO: `copy_members` will be deprecated in xoutil 1.8, use instead the same
# mechanisms as `xoutil.future`.
from xoutil.modules import copy_members as _copy_python_module_members

this = _copy_python_module_members(_odoo_expression)
del _copy_python_module_members
del _odoo_expression


try:
    from xoutil.objects import crossmethod  # TODO: migrate
except ImportError:

    class crossmethod(object):
        """Decorate a function as static or instance level.

        Example:

          >>> class Mule(object):
          ...     @crossmethod
          ...     def print_args(*args):
          ...         print(args)

          # Call it as a staticmethod
          >>> Mule.print_args()
          ()

          # Call it as an instance
          >>> Mule().print_args()   # doctest: +ELLIPSIS
          (<...Mule object at ...>,)

        .. note:: This is different from `hybridmethod`:func:.  Hybrid method
                  always receive the implicit argument (either `cls` or
                  `self`).

        """

        def __init__(self, func):
            self.func = func

        def __get__(self, instance, owner):
            if instance is None:
                return self.func
            else:
                return self.func.__get__(instance, owner)


UNARY_OPERATORS = [this.NOT_OPERATOR]
BINARY_OPERATORS = [this.AND_OPERATOR, this.OR_OPERATOR]
KIND_OPERATOR = "OPERATOR"
KIND_TERM = "TERM"


# Exports normalize_leaf so that we can replace 'from xoeuf.odoo.
def normalize_leaf(term):
    if this.is_leaf(term):
        left, operator, right = this.normalize_leaf(term)
        # Avoid no hashable values in domain terms
        if not getattr(right, "__hash__", False) and hasattr(right, "__iter__"):  # noqa
            right = tuple(right)
        return left, operator, right
    return term


class Domain(list):
    """A predicate expressed as an Odoo domain.

    .. note:: This is an *operational* wrapper around normal Odoo domains
              (lists) with methods to do logical manipulation of such values.

    It's a subtype of Odoo domains (i.e `list`:class:), which means that
    wherever an Odoo domain is expected you can use a Domain.  See the `Liskov
    substitution principle`__.

    __ https://en.wikipedia.org/wiki/Liskov_substitution_principle

    """

    def __init__(self, seq=None):
        # TODO: Can you do some sanity check to avoid common mistakes?  For
        # me, it's normal that I do ``Model.search(['field', '=', value])``
        # and forget the tuple...
        from xoeuf.odoo.tools.safe_eval import const_eval

        seq = seq or ()
        # some times the domains are saved in db in char or text fields.
        if isinstance(seq, string_types):
            seq = const_eval(seq)
        super(Domain, self).__init__(seq)

    def implies(self, other):
        """Check if a domain implies another.

        For any two domains `A` and `B`, the following rules are always true:

        - ``A.implies(A)``
        - ``(A & B).implies(A)``
        - ``A.implies(A | B)``.
        - ``not B.implies(A) == (A | B).implies(A)``

        """
        other = DomainTree(Domain(other).second_normal_form)
        return DomainTree(self.second_normal_form).implies(other)

    @property
    def first_normal_form(self):
        """The first normal form.

        The first normal form is the same domain with all `and` operators
        explicit.

        For instance, having ``domain`` value like::

            >>> domain = Domain(
            ...     [('field_y', 'not in', False), ('field_x', '!=', 'value')]
            ... )

        Then::

            >>> domain.first_normal_form
            ['&', ('field_y', 'not in', False), ('field_x', '!=', 'value')]

        """
        return Domain(this.normalize_domain(self))

    @property
    def second_normal_form(self):
        """The second normal form.

        After obtaining the `first_normal_from`:attr:, change terms to its
        canonical form, and distribute the `not` logical operator inside
        terms.

        .. seealso:: `normalize_leaf`:func: for the canonical form of terms.

        .. seealso:: `distribute_not`:meth:

        Example::

            >>> domain1 = Domain([
            ...     ('field_x', 'not in', False),
            ...     '!',
            ...     ('field_y', '>', 1)
            ... ])

            >>> domain2 = Domain([
            ...     '|',
            ...     ('field_x', 'not in', False),
            ...     ('field_y', '<>', 'value'),
            ...     ('field_z', '>', 1)
            ... ])

            >>> domain3 = Domain([
            ...     '|',
            ...     ('field_x', 'not in', False),
            ...     '!',
            ...     ('field_y', '<>', 'value'),
            ...     ('field_z', '>', 1)
            ... ])

            >>> domain1.first_normal_form
            ['&', ('field_x', '!=', False), ('field_y', '<=', 1)]

            >>> domain2.first_normal_form
            [
                '&',
                '|',
                ('field_x', '!=', False),
                ('field_y', '!=', 'value'),
                ('field_z', '>', 1)
            ]

            >>> domain3.first_normal_form
            [
                '&',
                '|',
                ('field_x', '!=', False),
                ('field_y', '=', 'value'),
                ('field_z', '>', 1)
            ]

        """
        res = self.first_normal_form
        res = Domain((normalize_leaf(item) for item in res))
        return res.distribute_not()

    @property
    def simplified(self):
        """A simplified second normal form of the domain.

        Examples:

            >>> domain1 = Domain([
            ...     ('field_x', '!=', False),
            ...     ('field_y', '=', 'value'),
            ...     ('field_z', 'in', (1, 2, 3)),
            ...     ('field_y', '=', 'value'),
            ... ])
            >>> domain1.simplified
            [
                '&',
                '&',
                ('field_x', '!=', False),
                ('field_y', '=', 'value'),
                ('field_z', 'in', (1, 2, 3))
            ]

            >>> domain2 = Domain([
            ...     '|',
            ...     ('field_x', '!=', False),
            ...     ('field_y', '=', 'value'),
            ...     ('field_z', 'in', (1, 2, 3)),
            ...     '|',
            ...     ('field_y', '=', 'value'),
            ...     ('field_x', '!=', False)
            ... ])
            >>> domain2.simplified
            [
                '&',
                '|',
                ('field_x', '!=', False),
                ('field_y', '=', 'value'),
                ('field_z', 'in', (1, 2, 3))
            ]

            >>> domain3 = Domain([
            ...     ('field_y', '!=', False),
            ...     ('field_x', 'in', (1,)),
            ...     '|',
            ...     ('field_y', '!=', False),
            ...     '|',
            ...     ('field_x', 'in', (1,)),
            ...     ('field_x', 'in', (2,))
            ... ])
            >>> domain3.simplified
            ['&', ('field_x', 'in', (1,)), ('field_y', '!=', False)]

        """
        return DomainTree(self.second_normal_form).get_simplified_domain()

    def distribute_not(self):
        """Return a new domain without `not` operators.

        For instance, having ``domain`` value like::

            >>> domain = Domain([
            ...     '!', ('field_x', '!=', False),
            ...     '!', ('field_y', '=', 'value'),
            ...     '!', ('field_z', 'in', (1, 2, 3)),
            ...     '!', ('field_w', '>', 1),
            ... ])

            >>> domain.distribute_not()
            [
                '&',
                '&',
                '&',
                ('field_x', '=', False),
                ('field_y', '!=', 'value'),
                ('field_z', 'not in', (1, 2, 3)),
                ('field_w', '<=', 1)
            ]

        """
        return Domain(this.distribute_not(self.first_normal_form))

    @crossmethod
    def AND(*domains):
        """Join given domains using `and` operator.

        :return: A domain if first normal form.

        """
        return Domain(
            this.AND([Domain(domain).second_normal_form for domain in domains])
        )

    __and__ = __rand__ = AND

    @crossmethod
    def OR(*domains):
        """Join given domains using `or` operator.

        :return: A domain if first normal form.

        """
        return Domain(
            this.OR([Domain(domain).second_normal_form for domain in domains])
        )

    __or__ = __ror__ = OR

    def __invert__(self):
        return Domain(["!"] + self.second_normal_form)

    def __eq__(self, other):
        """Two domains are equivalent if both have similar DomainTree.

        """
        # TODO: In logic, we can identify two predicates if: a implies
        # b and b implies a, although this has nothing to do with them being
        # the *same* predicate.  However, since this implementation only
        # yields True when both domains have the same hash, we can find a, b
        # such that a implies b and b implies a, but a != b.
        other = Domain(other)
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(DomainTree(self.second_normal_form))

    def asfilter(self, this="this"):
        """Return a callable which is equivalent to the domain.

        This method translates the domain to a `ast.Lambda <ast>`:mod: and
        compiles it.

        The main property is that is ``res`` is the result of a
        ``search(domain)``; then filtering by ``domain.asfilter()`` does not
        filter-out any record.

        The domain cannot use 'child_of', '=like' or '=ilike'.

        .. note:: In Python ``0 == False``, so Odoo treats 0 specially in the
                  context of 'not in' and 'in'.  See `PR 31408`__ for more
                  information.

        __ https://github.com/odoo/odoo/pull/31408

        :param this: The name of the argument in the lambda.  All attributes
            in the domain are get from this argument.

        The lambda created for::

            Domain([('state', 'in', ('draft', 'open'))]).asfilter()

        is equivalent to::

            lambda this: this.state in ('draft', 'open')

        .. rubric:: Traversing fields in domains

        If your domain uses field traversal (e.g ``('line_ids.state', ...)``)
        the generated lambda will use ``mapped()`` and ``filtered()`` instead
        of simple ``ast.Attribute`` nodes.  Thus the lamda for::

          Domain([('order_id.line_ids.state', '=', 'open')]).asfilter(this='t')

        is equivalent to::

          lambda t: (t.mapped('order_id.line_ids')
                      .filtered(lambda r: r.state == 'open'))

        .. versionadded:: 0.54.0

        .. versionchanged:: 0.55.0 Change the behavior of fields traversal, so
           that filters over x2many fields work.

        """
        return eval(compile(self._get_filter_ast(this), "<domain>", "eval"))

    def _get_filter_ast(self, this="this"):
        return DomainTree(self.second_normal_form)._get_filter_ast(this)


class DomainTerm(object):
    def __init__(self, term):
        if isinstance(term, DomainTerm):
            term = term.original
        self.original = term
        term = normalize_leaf(term)
        self.normalized = term
        if this.is_operator(term):
            self.is_operator = True
            self.operator = term
            self.is_leaf = self.left = self.right = False
        elif this.is_leaf(term):
            self.is_operator = False
            self.is_leaf = True
            self.left, self.operator, self.right = term
        else:
            # TODO: May be a # TypeError.
            raise ValueError("Invalid domain term %r" % term)

    def __getitem__(self, x):
        if self.is_leaf:
            return (self.left, self.operator, self.right)[x]
        else:
            return self.operator[x]

    def __eq__(self, other):
        if not isinstance(other, DomainTerm):
            other = DomainTerm(other)
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        if self.original == self.normalized:
            return repr(self.original)
        else:
            return "%r => %r" % (self.original, self.normalized)

    operators_implication = {
        "=?": lambda x, y: operator.eq(x, y) or y is False,
        ">": operator.ge,
        ">=": operator.ge,
        "<": operator.le,
        "<=": operator.le,
        "in": lambda x, y: all(i in y for i in x),
        "not in": lambda x, y: all(i in x for i in y),
        "like": lambda x, y: x.find(y) >= 0,
        # TODO: asd_g imply asdfg `_` == any character
        "=like": lambda x, y: x.find(y) >= 0,
        "not like": lambda x, y: y.find(x) >= 0,
        "ilike": lambda x, y: x.lower().find(y.lower()) >= 0,
        # TODO: asd_g imply asdfg `_` == any character
        "=ilike": lambda x, y: x.lower().find(y.lower()) >= 0,
        "not ilike": lambda x, y: y.lower().find(x.lower()) >= 0,
    }

    def implies(self, other):
        other = DomainTerm(other)
        # equals terms are implied.
        if self == other:
            return True
        # terms which different left operand can't be compared.
        elif self.left != other.left:
            return False
        elif self.is_operator:
            # & => &
            return self.operator == other.operator
        else:
            # (x = 1)  => (x = 1)
            # (x > 1) => (x > 0)
            # (x in (1,2,3)) => (x in (1,2,3,4))
            if self.operator == other.operator:
                compare = self.operators_implication.get(other.operator)
                return compare(self.right, other.right) if compare else False
            else:
                # TODO: x = 1  implies x != 2; x = 2 1 implies x > 1
                return False

    def __hash__(self):
        return hash(self.normalized)


class DomainTree(object):
    """Tree structure to express Odoo domains.

    .. warning:: This class in not a public API of this model.  Its attributes
       may change in incompatible ways from one release to the other.

    A domain like this::

      [
          '&',
          '&',
          ('field_y', '!=', False),
          ('field_x', '=', 'value'),
          '|',
          ('field_z', 'in', (1, 2, 3)),
          ('field_w', '>', 1)
      ]

    Is represented like this::

                                     +-----+
                                     | '&' |
                                     +--+--+
                                        |
                   +--------------------+------------------------+
                   |                    |                        |
      +------------+-----------+ +------+------------------+  +--+--+
      |('field_y', '!=', False)| |('field_x', '=', 'value')|  | '|' |
      +------------------------+ +-------------------------+  +--+--+
                                                                 |
                                  +------------------------------+--------+
                                  |                                       |
                     +------------+---------------+       +---------------+---+
                     |('field_z', 'in', (1, 2, 3))|       |('field_w', '>', 1)|
                     +----------------------------+       +-------------------+

    .. warning:: The domain must be in the second normal form.

    """

    def __init__(self, domain, parent=None):
        term = domain.pop(0)
        self.term = DomainTerm(term)
        self.parent = parent
        if term in this.DOMAIN_OPERATORS:
            count = 2  # minimum number of operand in an operation.
            children = set()
            while count:
                if domain[0] == term:
                    count += 1
                    domain.pop(0)
                else:
                    child = DomainTree(domain, self)
                    # A & ((B & C) | A) should be simplified as A & B & C
                    if child.term == self.term:
                        children |= child.children
                    else:
                        children.add(child)
                    count -= 1
            # if a tree node have only one child it be come into it child.
            if len(children) == 1:
                child = children.pop()
                self.term = child.term
                self.children = child.children
            else:
                self.children = children
        else:
            self.children = set()
        self._simplify()

    @property
    def is_operator(self):
        return self.term in this.DOMAIN_OPERATORS

    @property
    def is_leaf(self):
        return not self.is_operator

    def _simplify(self):
        """Remove redundant branches.

        """
        for child in set(self.children):
            if self.term == this.AND_OPERATOR:
                # If current `child` is implied by any other ignore it.
                func = lambda x, y: y.implies(x)
            else:
                # If current `child` implies any other ignore it.
                func = lambda x, y: x.implies(y)
            if any(func(child, y) for y in self.children - {child}):
                self.children.remove(child)
        if len(self.children) == 1:
            _self = self.children.pop()
            self.children = _self.children
            self.term = _self.term

    @property
    def sorted_children(self):
        # TODO: Sort by hash is weird.  What does it mean?
        return sorted(self.children, key=lambda item: hash(item))

    def get_simplified_domain(self):
        if self.parent:
            res = Domain(self.term.original for x in range(1, len(self.children) or 2))
        elif self.is_leaf:
            res = Domain([self.term.original])
        else:
            # Initials `&` aren't needed.
            res = Domain([self.term.original] if self.term == this.OR_OPERATOR else [])
        if not self.is_leaf:
            res.extend(
                chain(*(x.get_simplified_domain() for x in self.sorted_children))
            )
        return res

    def __repr__(self):
        if self.is_leaf:
            return repr(self.term)
        else:
            return "(%s)" % (" %r " % self.term).join(
                repr(child) for child in self.sorted_children
            )

    def __eq__(self, other):
        if self.is_leaf == other.is_leaf:
            if self.is_leaf:
                return self.term == other.term
            else:
                return (
                    self.term == other.term and not self.children ^ other.children
                )  # noqa
        return False

    def __ne__(self, other):
        return not self == other

    def implies(self, other):
        funct = all if other.term == this.AND_OPERATOR else any
        if self.is_leaf:
            # A => A
            if self.term.implies(other.term):
                return True
            # A => A | B
            if other.is_operator and funct(
                self.implies(child) for child in other.sorted_children
            ):
                return True
        elif self.is_operator:
            funct2 = any if self.term == this.AND_OPERATOR else all
            # A & B => A
            if funct2(child.implies(other) for child in self.sorted_children):
                return True
            # A & B => A | B
            if funct(
                funct2(child.implies(other_child) for child in self.sorted_children)
                for other_child in other.sorted_children
            ):
                return True
        return False

    def __hash__(self):
        return hash(tuple([self.term] + self.sorted_children))

    def _get_filter_ast(self, this="this"):
        "Get compilable AST of the lambda obtained by `get_filter`:func:."
        stack = []
        for kind, term in self.walk():
            if kind == KIND_TERM:
                fieldname, op, value = term
                constructor = _TERM_CONSTRUCTOR[op]
                stack.append(constructor(this, fieldname, value))
            else:
                assert kind == KIND_OPERATOR
                if term in BINARY_OPERATORS:
                    args = [stack.pop(), stack.pop()]
                else:
                    args = [stack.pop()]
                constructor = _TERM_CONSTRUCTOR[term]
                stack.append(constructor(*args))
        node = stack.pop()
        assert not stack, "Remaining nodes in the stack: {}".format(stack)
        fn = ql.ensure_compilable(
            ql.Expression(ql.Lambda(ql.make_arguments(this), node))
        )
        return fn

    def walk(self):
        """Performs a post-fix walk of the tree.

        Yields tuples of ``(kind, what)``.  `kind` can be either 'TERM' or
        'OPERATOR'.  `what` will be the term or operator.  For `term` it will
        the tuple ``(field, operator, arg)`` in Odoo domains.  For `operator`
        it will be the string identifying the operator.

        Despite that the tree allows AND (``&``) and OR (``|``) operator to
        have more than two children, we generate OPERATOR items for the
        equivalent binary expression tree::

           >>> from xoeuf.osv.expression import Domain, DomainTree
           >>> d = Domain([('a', '=', 1), ('b', '=', 2), ('c', '=', 3)])
           >>> tree = DomainTree(d.second_normal_form)
           >>> list(tree.walk())
           [('TERM', ('a', '=', 1)),
            ('TERM', ('c', '=', 3)),
            ('TERM', ('b', '=', 2)),
            ('OPERATOR', '&'),
            ('OPERATOR', '&')]

        .. note:: DomainTree always simplify its input.  The walk is not a
           syntactical walk of the input but of its simplified form.  Since
           all operators are commutative, the order of the children may differ
           widely from run to the other.

        """
        if self.is_leaf:
            yield (KIND_TERM, self.term.original)
        else:
            for which in self.children:
                for what in which.walk():
                    yield what
            yield (KIND_OPERATOR, self.term)
            # Since the tree groups more than two children under a single
            # instance of binary operator, we have to yield it once per pair.
            if self.term.original in BINARY_OPERATORS:
                for _ in range(len(self.children) - 2):
                    yield (KIND_OPERATOR, self.term)


# Exports AND and OR so that we can replace 'from xoeuf.odoo.
def AND(domains):
    return Domain.AND(*domains)


def OR(domains):
    return Domain.OR(*domains)


def _constructor_not(node):
    return ql.UnaryOp(ql.Not(), node)


def _constructor_and(*operands):
    return ql.BoolOp(ql.And(), list(operands))


def _constructor_or(*operands):
    return ql.BoolOp(ql.Or(), list(operands))


def _get_mapped(node, fieldname):
    attrs, field = fieldname.rsplit(".", 1)
    mapped_fn = ql.make_attr(node, "mapped")
    return (ql.make_call(mapped_fn, _constructor_from_value(attrs)), field)


def _constructor_getattr(node, fieldname):
    if isinstance(fieldname, string_types):
        if "." in fieldname:
            result = _get_mapped(node, fieldname)
        else:
            result = ql.make_attr(node, fieldname)
    else:
        result = _constructor_from_value(fieldname)
    return result


def _get_constructor(qst):
    """Return a constructor for AST for a term `(fielname, <op>, value)`.

    :param qst: Any of the operators AST classes available for comparisons.

    The result is a function that with signaure ``(this, fieldname, value)``.

    :param this: The AST node to which the operation is applied.

    :param fieldname: The name of the field being operated upon.

    :param value: The second operand of the operator.

    If the fieldname doesn't contain a dot '.'; return a simple AST.  A domain
    like::

        Domain(['state', '=', 'open']).asfilter()

    yields a callable equivalent to::

        lambda this: this.state == 'open'

    If the fieldname contains a dot '.'; return an AST using ``mapped()`` and
    ``filtered()`` as appropriate:

        Domain(['parent_id.children_ids.state', '=', 'open']).asfilter()

    yields a callable equivalent to::

        lambda this: this.mapped('parent_id.children_ids').filtered(
            lambda r: r.state == 'open'
        )

    """

    def result(this, fieldname, value):
        node = _constructor_getattr(ql.Name(this, ql.Load()), fieldname)
        if not isinstance(node, tuple):
            return ql.Compare(node, [qst()], [_constructor_from_value(value)])
        else:
            # node.filtered(lambda x: x.field <CMP> <value>)
            mapped, field = node
            lambda_ast = ql.Lambda(
                ql.make_arguments("x"),
                ql.Compare(
                    ql.make_attr(ql.Name("x", ql.Load()), field),
                    [qst()],
                    [_constructor_from_value(value)],
                ),
            )
            filtered_fn = ql.make_attr(mapped, "filtered")
            return ql.make_call(filtered_fn, lambda_ast)

    return result


_constructor_eq = _get_constructor(ql.Eq)
_constructor_ne = _get_constructor(ql.NotEq)
_constructor_le = _get_constructor(ql.LtE)
_constructor_lt = _get_constructor(ql.Lt)
_constructor_ge = _get_constructor(ql.GtE)
_constructor_gt = _get_constructor(ql.Gt)


def _constructor_in(this, fieldname, value, qst=ql.In):
    """Create the AST for a term `(fielname, '[not ]in', value)`.

    The difference with `standard <_get_constructor>`:func: constructors is
    that, to comply with Odoo's interpretation of 'in' and 'not in', this
    function removes any 0 or False in `value`.

    """
    # Filtering False is the same Odoo does; which causes 0 to be removed
    # also.  See https://github.com/odoo/odoo/pull/31408
    assert qst in (ql.In, ql.NotIn)
    value = [x for x in value if x != False]  # noqa
    return _get_constructor(qst)(this, fieldname, value)


def _constructor_not_in(this, fieldname, value):
    return _constructor_in(this, fieldname, value, qst=ql.NotIn)


def _constructor_like(this, fieldname, value, qst=ql.In):
    """Create the AST for a term `(fielname, '[not ]like', value)`.

    We use 'in' ('not in') Python operators; so the procedure AST are those
    matching syntax 'value in this.fieldname' when fieldname doesn't contain a
    dot.  If `fieldname` contains a dot the result changes to::

       headattrs, lastattr = fiedname.rsplit('.', 1)
       this.mapped(headattrs).filtered(lambda r: value in getattr(r, lastattr))

    """
    assert qst in (ql.In, ql.NotIn)
    node = _constructor_getattr(ql.Name(this, ql.Load()), fieldname)
    if isinstance(node, tuple):
        # this.mapped(attrs).filtered(lambda r: value in r.field)
        mapped, field = node
        lambda_arg = ql.Lambda(
            ql.make_arguments("r"),
            ql.Compare(
                _constructor_from_value(value),
                [qst()],
                [ql.make_attr(ql.Name("r", ql.Load()), field)],
            ),
        )
        fn = ql.make_attr(mapped, "filtered")
        return ql.make_call(fn, lambda_arg)
    else:
        # ``value in this.fieldname``
        return ql.Compare(_constructor_from_value(value), [qst()], [node])


def _constructor_not_like(this, fieldname, value):
    return _constructor_like(this, fieldname, value, qst=ql.NotIn)


def _constructor_ilike(this, fieldname, value, qst=ql.In):
    assert qst in (ql.In, ql.NotIn)
    node = _constructor_getattr(ql.Name(this, ql.Load()), fieldname)
    if isinstance(node, tuple):
        # this.mapped(attrs).filtered(lambda r: value.lower() in r.field.lower())``
        mapped, field = node
        lambda_arg = ql.Lambda(
            ql.make_arguments("r"),
            ql.Compare(
                ql.make_argless_call(
                    _constructor_getattr(_constructor_from_value(value), "lower")
                ),
                [qst()],
                [
                    ql.make_argless_call(
                        ql.make_attr(
                            ql.make_attr(ql.Name("r", ql.Load()), field), "lower"
                        )
                    )
                ],
            ),
        )
        fn = ql.make_attr(mapped, "filtered")
        return ql.make_call(fn, lambda_arg)
    else:
        # ``value.lower() in this.fieldname.lower()``
        node = _constructor_getattr(node, "lower")
        fn = ql.make_argless_call(node)
        return ql.Compare(
            ql.make_argless_call(
                _constructor_getattr(_constructor_from_value(value), "lower")
            ),
            [qst()],
            [fn],
        )


def _constructor_not_ilike(this, fieldname, value):
    return _constructor_ilike(this, fieldname, value, qst=ql.NotIn)


def _constructor_from_value(value):
    expr = ql.parse(repr(value))
    return expr.body


_TERM_CONSTRUCTOR = {
    "!": _constructor_not,
    "&": _constructor_and,
    "|": _constructor_or,
    "=": _constructor_eq,
    "==": _constructor_eq,
    "=?": _constructor_eq,
    "!=": _constructor_ne,
    "<>": _constructor_ne,
    "<=": _constructor_le,
    "<": _constructor_lt,
    ">=": _constructor_ge,
    ">": _constructor_gt,
    "in": _constructor_in,
    "not in": _constructor_not_in,
    "like": _constructor_like,
    "ilike": _constructor_ilike,
    "not like": _constructor_not_like,
    "not ilike": _constructor_not_ilike,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import unittest


class TestModules(unittest.TestCase):
    def test_get_object_module(self):
        from .. import Foo
        from xoeuf.modules import get_object_module

        self.assertEqual(get_object_module(Foo), Foo.module)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)


from xoeuf.odoo import api, fields


@api.model
def _tz_get(self):
    import pytz

    def key(tz):
        return tz if not tz.startswith("Etc/") else "_"

    return [(tz, "(" + tz + ")") for tz in sorted(pytz.all_timezones, key=key)]


def TimezoneSelection(*args, **kwargs):
    """A selection field for installed timezones.

    """
    return fields.Selection(_tz_get, *args, **kwargs)
